import abc
from typing import List


class Dataset:
    TRAIN = "train"
    TEST = "test"
    SUBSAMPLED = "subsampled"
    PREPROCESSED = "preprocessed"

    def __init__(self, encoded_data=None, name: str = None, identifier: str = None, labels: dict = None, example_weights: list = None):
        self.encoded_data = encoded_data
        self.identifier = identifier
        self.name = name if name is not None else self.identifier
        self.labels = labels
        self.example_weights = example_weights

    @classmethod
    @abc.abstractmethod
    def build_from_objects(cls, **kwargs):
        pass

    @abc.abstractmethod
    def make_subset(self, example_indices, path, dataset_type: str):
        pass

    @abc.abstractmethod
    def get_attribute(self, attribute: str, as_list: bool = True):
        pass

    @abc.abstractmethod
    def get_attributes(self, attributes: List[str], as_list: bool = True):
        pass

    @abc.abstractmethod
    def get_example_count(self):
        pass

    @abc.abstractmethod
    def get_data(self, batch_size: int = 1):
        pass

    @abc.abstractmethod
    def get_batch(self, batch_size: int = 1):
        pass

    @abc.abstractmethod
    def get_example_ids(self):
        pass

    @abc.abstractmethod
    def get_label_names(self):
        pass

    @abc.abstractmethod
    def clone(self, keep_identifier: bool = False):
        pass

    @abc.abstractmethod
    def get_metadata(self, field_names: list, return_df: bool = False):
        pass

    @abc.abstractmethod
    def get_data_from_index_range(self, start_index: int, end_index: int):
        pass

    def set_example_weights(self, example_weights: list):
        if example_weights is not None:
            assert len(example_weights) == self.get_example_count(), f"{self.__class__.__name__}: trying to set example weights " \
                                                                 f"for dataset {self.identifier} but number of weights ({len(example_weights)}) " \
                                                                 f"does not match example count ({self.get_example_count()}). "
        self.example_weights = example_weights

    def get_example_weights(self):
        return self.example_weights


import pandas as pd

from immuneML.IO.dataset_import.DataImport import DataImport
from immuneML.IO.dataset_import.DatasetImportParams import DatasetImportParams
from immuneML.data_model.dataset import Dataset
from immuneML.data_model.receptor.ChainPair import ChainPair
from immuneML.data_model.receptor.RegionType import RegionType
from immuneML.data_model.repertoire.Repertoire import Repertoire
from immuneML.util.ImportHelper import ImportHelper
from scripts.specification_util import update_docs_per_mapping


class GenericImport(DataImport):
    """
    Imports data from any tabular file into a Repertoire-, Sequence- or ReceptorDataset. RepertoireDatasets should be
    used when making predictions per repertoire, such as predicting a disease state. SequenceDatasets or ReceptorDatasets
    should be used when predicting values for unpaired (single-chain) and paired immune receptors respectively,
    like antigen specificity.

    This importer works similarly to other importers, but has no predefined default values for which fields are imported,
    and can therefore be tailored to import data from various different tabular files with headers.

    For ReceptorDatasets: this importer assumes the two receptor sequences appear on different lines in the file, and can
    be paired together by a common sequence identifier. If you instead want to import a ReceptorDataset from a tabular
    file that contains both receptor chains on one line, see :ref:`SingleLineReceptor` import


    Specification arguments:

    - path (str): For RepertoireDatasets, this is the path to a directory with files to import. For Sequence- or ReceptorDatasets this path may either be the path to the file to import, or the path to the folder locating one or multiple files with .tsv, .csv or .txt extensions. By default path is set to the current working directory.

    - is_repertoire (bool): If True, this imports a RepertoireDataset. If False, it imports a SequenceDataset or ReceptorDataset. By default, is_repertoire is set to True.

    - metadata_file (str): Required for RepertoireDatasets. This parameter specifies the path to the metadata file. This is a csv file with columns filename, subject_id and arbitrary other columns which can be used as labels in instructions. For setting Sequence- or ReceptorDataset metadata, metadata_file is ignored, see metadata_column_mapping instead.

    - paired (str): Required for Sequence- or ReceptorDatasets. This parameter determines whether to import a SequenceDataset (paired = False) or a ReceptorDataset (paired = True). In a ReceptorDataset, two sequences with chain types specified by receptor_chains are paired together based on a common identifier. This identifier should be mapped to the immuneML field 'sequence_identifiers' using the column_mapping.

    - receptor_chains (str): Required for ReceptorDatasets. Determines which pair of chains to import for each Receptor. Valid values for receptor_chains are the names of the ChainPair enum.

    - import_illegal_characters (bool): Whether to import sequences that contain illegal characters, i.e., characters that do not appear in the sequence alphabet (amino acids including stop codon '*', or nucleotides). When set to false, filtering is only applied to the sequence type of interest (when running immuneML in amino acid mode, only entries with illegal characters in the amino acid sequence are removed). By default import_illegal_characters is False.

    - import_empty_nt_sequences (bool): imports sequences which have an empty nucleotide sequence field; can be True or False. By default, import_empty_nt_sequences is set to True.

    - import_empty_aa_sequences (bool): imports sequences which have an empty amino acid sequence field; can be True or False; for analysis on amino acid sequences, this parameter should be False (import only non-empty amino acid sequences). By default, import_empty_aa_sequences is set to False.

    - region_type (str): Which part of the sequence to import. By default, this value is set to IMGT_CDR3. This means immuneML assumes the IMGT junction (including leading C and trailing Y/F amino acids) is used in the input file, and the first and last amino acids will be removed from the sequences to retrieve the IMGT CDR3 sequence. Specifying any other value will result in importing the sequences as they are. Valid values for region_type are the names of the :py:obj:`~immuneML.data_model.receptor.RegionType.RegionType` enum.

    - column_mapping (dict): Required for all datasets. A mapping where the keys are the column names in the input file, and the values correspond to the names used in immuneML's internal data representation. Valid immuneML fields that can be specified here are defined by Repertoire.FIELDS. At least sequences (nucleotide) or sequence_aas (amino acids) must be specified, but all other fields are optional. A column mapping can look for example like this:

        .. indent with spaces
        .. code-block:: yaml

            file_column_amino_acids: sequence_aa
            file_column_v_genes: v_call
            file_column_j_genes: j_call
            file_column_frequencies: duplicate_count

    - column_mapping_synonyms (dict): This is a column mapping that can be used if a column could have alternative names. The formatting is the same as column_mapping. If some columns specified in column_mapping are not found in the file, the columns specified in column_mapping_synonyms are instead attempted to be loaded. For Generic import, there is no default column_mapping_synonyms.

    - metadata_column_mapping (dict): Specifies metadata for Sequence- and ReceptorDatasets. This should specify a mapping similar to column_mapping where keys are file column names and values are the names that are internally used in immuneML as metadata fields. These metadata fields can be used as prediction labels for Sequence- and ReceptorDatasets. This parameter can also be used to specify sequence-level metadata columns for RepertoireDatasets, which can be used by reports. To set prediction label metadata for RepertoireDatasets, see metadata_file instead. There is no default metadata_column_mapping.

        .. indent with spaces
        .. code-block:: yaml

                file_column_antigen_specificity: antigen_specificity

    - columns_to_load (list): Optional; specifies which columns to load from the input file. This may be useful if the input files contain many unused columns. If no value is specified, all columns are loaded.

    - separator (str): Required parameter. Column separator, for example "\\t" or ",". The default value is "\\t"


    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_generic_dataset:
            format: Generic
            params:
                path: path/to/files/
                is_repertoire: True # whether to import a RepertoireDataset
                metadata_file: path/to/metadata.csv # metadata file for RepertoireDataset
                paired: False # whether to import SequenceDataset (False) or ReceptorDataset (True) when is_repertoire = False
                receptor_chains: TRA_TRB # what chain pair to import for a ReceptorDataset
                separator: "\\t" # column separator
                import_illegal_characters: False # remove sequences with illegal characters for the sequence_type being used
                import_empty_nt_sequences: True # keep sequences even though the nucleotide sequence might be empty
                import_empty_aa_sequences: False # filter out sequences if they don't have sequence_aa set
                region_type: IMGT_CDR3 # what part of the sequence to import
                column_mapping: # column mapping file: immuneML
                    file_column_amino_acids: sequence_aas
                    file_column_v_genes: v_call
                    file_column_j_genes: j_call
                    file_column_frequencies: duplicate_count
                metadata_column_mapping: # metadata column mapping file: immuneML
                    file_column_antigen_specificity: antigen_specificity
                columns_to_load:  # which subset of columns to load from the file
                    - file_column_amino_acids
                    - file_column_v_genes
                    - file_column_j_genes
                    - file_column_frequencies
                    - file_column_antigen_specificity

    """

    @staticmethod
    def import_dataset(params: dict, dataset_name: str) -> Dataset:
        return ImportHelper.import_dataset(GenericImport, params, dataset_name)

    @staticmethod
    def preprocess_dataframe(df: pd.DataFrame, params: DatasetImportParams):
        ImportHelper.drop_empty_sequences(df, params.import_empty_aa_sequences, params.import_empty_nt_sequences)
        ImportHelper.drop_illegal_character_sequences(df, params.import_illegal_characters, params.import_with_stop_codon)
        ImportHelper.junction_to_cdr3(df, params.region_type)
        df.loc[:, "region_type"] = params.region_type.name
        ImportHelper.load_chains(df)

        return df

    @staticmethod
    def import_receptors(df, params):
        df["receptor_id"] = df["sequence_id"]
        return ImportHelper.import_receptors(df, params)

    @staticmethod
    def get_documentation():
        doc = str(GenericImport.__doc__)

        chain_pair_values = str([chain_pair.name for chain_pair in ChainPair])[1:-1].replace("'", "`")
        region_type_values = str([region_type.name for region_type in RegionType])[1:-1].replace("'", "`")
        repertoire_fields = list(Repertoire.FIELDS)
        repertoire_fields.remove("region_type")

        mapping = {
            "Valid values for receptor_chains are the names of the ChainPair enum.": f"Valid values are {chain_pair_values}.",
            "Valid values for region_type are the names of the :py:obj:`~immuneML.data_model.receptor.RegionType.RegionType` enum.": f"Valid values are {region_type_values}.",
            "Valid immuneML fields that can be specified here are defined by Repertoire.FIELDS": f"Valid immuneML fields that can be specified here are {repertoire_fields}."
        }
        doc = update_docs_per_mapping(doc, mapping)
        return doc



import warnings
from pathlib import Path
from typing import List

import pickle
import pandas as pd
import plotly.express as px
import numpy as np

from immuneML.data_model.dataset.RepertoireDataset import RepertoireDataset
from immuneML.dsl.instruction_parsers.LabelHelper import LabelHelper
from immuneML.encodings.abundance_encoding.AbundanceEncoderHelper import AbundanceEncoderHelper
from immuneML.encodings.abundance_encoding.CompAIRRSequenceAbundanceEncoder import CompAIRRSequenceAbundanceEncoder
from immuneML.encodings.abundance_encoding.KmerAbundanceEncoder import KmerAbundanceEncoder
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.SignificantFeaturesHelper import SignificantFeaturesHelper


class SignificantFeatures(DataReport):
    """
    Plots a boxplot of the number of significant features (label-associated k-mers or sequences) per Repertoire according to Fisher's exact test,
    across different classes for the given label.

    Internally uses the :py:obj:`~immuneML.encodings.abundance_encoding.KmerAbundanceEncoder.KmerAbundanceEncoder` for calculating
    significant k-mers, and
    :py:obj:`~immuneML.encodings.abundance_encoding.SequenceAbundanceEncoder.SequenceAbundanceEncoder` or
    :py:obj:`~immuneML.encodings.abundance_encoding.CompAIRRSequenceAbundanceEncoder.CompAIRRSequenceAbundanceEncoder`
    to calculate significant full sequences (depending on whether the argument compairr_path was set).

    Specification arguments:

    - p_values (list): The p value thresholds to be used by Fisher's exact test. Each p-value specified here will become
      one panel in the output figure.

    - k_values (list): Length of the k-mers (number of amino acids) created by the
      :py:obj:`~immuneML.encodings.abundance_encoding.KmerAbundanceEncoder.KmerAbundanceEncoder`.
      When using a full sequence encoding (:py:obj:`~immuneML.encodings.abundance_encoding.SequenceAbundanceEncoder.SequenceAbundanceEncoder` or
      :py:obj:`~immuneML.encodings.abundance_encoding.CompAIRRSequenceAbundanceEncoder.CompAIRRSequenceAbundanceEncoder`), specify 'full_sequence' here.
      Each value specified under k_values will represent one boxplot in the output figure.

    - label (dict): A label configuration. One label should be specified, and the positive_class for this label should be defined. See the YAML specification below for an example.

    - compairr_path (str): If 'full_sequence' is listed under k_values, the path to the CompAIRR executable may be provided.
      If the compairr_path is specified, the :py:obj:`~immuneML.encodings.abundance_encoding.CompAIRRSequenceAbundanceEncoder.CompAIRRSequenceAbundanceEncoder`
      will be used to compute the significant sequences. If the path is not specified and 'full_sequence' is listed under
      k-values, :py:obj:`~immuneML.encodings.abundance_encoding.SequenceAbundanceEncoder.SequenceAbundanceEncoder` will be used.

    - log_scale (bool): Whether to plot the y axis in log10 scale (log_scale = True) or continuous scale (log_scale = False). By default, log_scale is False.


    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_significant_features_report:
            SignificantFeatures:
                p_values:
                    - 0.1
                    - 0.01
                    - 0.001
                    - 0.0001
                k_values:
                    - 3
                    - 4
                    - 5
                    - full_sequence
                compairr_path: path/to/compairr # can be specified if 'full_sequence' is listed under k_values
                label: # Define a label, and the positive class for that given label
                    CMV:
                        positive_class: +
                log_scale: False

    """

    @classmethod
    def build_object(cls, **kwargs):
        location = SignificantFeatures.__name__

        kwargs = SignificantFeaturesHelper.parse_parameters(kwargs, location)

        ParameterValidator.assert_keys_present(kwargs.keys(), ["log_scale"], location, location)
        ParameterValidator.assert_type_and_value(kwargs["log_scale"], bool, "SignificantFeatures", "log_scale")

        return SignificantFeatures(**kwargs)

    def __init__(self, dataset: RepertoireDataset = None, p_values: List[float] = None, k_values: List[int] = None,
                 label: dict = None, compairr_path: Path = None, log_scale: bool = False, result_path: Path = None,
                 name: str = None, number_of_processes: int = 1):
        super().__init__(dataset=dataset, result_path=result_path, number_of_processes=number_of_processes, name=name)
        self.p_values = p_values
        self.k_values = k_values
        self.label = label
        self.label_config = None
        self.compairr_path = compairr_path
        self.log_scale = log_scale

    def check_prerequisites(self):
        if isinstance(self.dataset, RepertoireDataset):
            return True
        else:
            warnings.warn(f"{SignificantFeatures.__name__}: report can be generated only from RepertoireDataset. Skipping this report...")
            return False

    def _generate(self) -> ReportResult:
        self.label_config = LabelHelper.create_label_config([self.label], self.dataset, SignificantFeatures.__name__, f"{SignificantFeatures.__name__}/label")

        plotting_data = self._compute_plotting_data()
        table_result = self._write_results_table(plotting_data)

        report_output_fig = self._safe_plot(plotting_data=plotting_data)
        output_figures = None if report_output_fig is None else [report_output_fig]

        return ReportResult(name=self.name,
                            info="The number of significant features (label-associated k-mers or sequences) per Repertoire according to Fisher's exact test, across different classes for the given label.",
                            output_figures=output_figures,
                            output_tables=[table_result])

    def _compute_plotting_data(self):
        result = {"encoding": [],
                  "p-value": [],
                  "class": [],
                  "significant_features": []}

        positive_class, negative_class = self._get_positive_negative_classes()

        for k in self.k_values:
            encoder_name = SignificantFeaturesHelper._get_encoder_name(k)

            for p_value in self.p_values:
                pos_class_feature_counts, neg_class_feature_counts = self._compute_significant_feature_counts(k, p_value)
                n_examples = len(pos_class_feature_counts) + len(neg_class_feature_counts)

                result["encoding"].extend([encoder_name] * n_examples)
                result["p-value"].extend([p_value] * n_examples)
                result["class"].extend(
                    [positive_class] * len(pos_class_feature_counts) + [negative_class] * len(neg_class_feature_counts))
                result["significant_features"].extend(list(pos_class_feature_counts) + list(neg_class_feature_counts))

        return pd.DataFrame(result)

    def _get_positive_negative_classes(self):
        label = self.label_config.get_label_objects()[0]
        positive_class = label.positive_class
        negative_class = label.get_binary_negative_class()

        return positive_class, negative_class

    def _get_encoder_result_path(self, k, p_value):
        result_path = self.result_path / f"{SignificantFeaturesHelper._get_encoder_name(k)}_{p_value}"
        PathBuilder.build(result_path)
        return result_path

    def _write_results_table(self, data) -> ReportOutput:
        table_path = self.result_path / f"significant_features_report.csv"
        data.to_csv(table_path, index=False)
        return ReportOutput(table_path, "Significant features across different Repertoire classes")

    def _plot(self, plotting_data):
        figure = px.box(plotting_data, x="encoding", y="significant_features", color="class",
                        facet_row=None, facet_col="p-value", log_y=self.log_scale,
                        labels={
                            "significant_features": "Number of significant features per AIRR according to Fisher's exact test",
                            "encoding": "Encoding",
                            "class": "Repertoire class"
                        }, template='plotly_white',
                        color_discrete_sequence=px.colors.diverging.Tealrose)

        file_path = self.result_path / f"significant_features_figure.html"

        figure.write_html(str(file_path))

        return ReportOutput(file_path, name="Significant features across different Repertoire classes")

    def _compute_significant_feature_counts(self, k, p_value):
        encoder_result_path = self._get_encoder_result_path(k, p_value)
        encoder_params = SignificantFeaturesHelper._build_encoder_params(self.label_config, encoder_result_path)

        if type(k) == int:
            pos_class_feature_counts, neg_class_feature_counts = self._compute_significant_kmer_counts(k, p_value, encoder_params)
        elif self.compairr_path is None:
            pos_class_feature_counts, neg_class_feature_counts = self._compute_significant_sequence_counts(p_value, encoder_params)
        else:
            pos_class_feature_counts, neg_class_feature_counts = self._compute_significant_compairr_sequence_counts(p_value, encoder_params)

        return pos_class_feature_counts, neg_class_feature_counts

    def _compute_significant_kmer_counts(self, k, p_value, encoder_params):
        encoder = SignificantFeaturesHelper._build_kmer_encoder(self.dataset, k, p_value, encoder_params)

        with encoder.relevant_indices_path.open("rb") as file:
            relevant_indices = pickle.load(file)

        relevant_feature_presence = np.sum(encoder.kmer_presence_matrix[relevant_indices], axis=0)

        return self._get_positive_negative_class(relevant_feature_presence, encoder.matrix_repertoire_ids)

    def _compute_significant_sequence_counts(self, p_value, encoder_params):
        encoder = SignificantFeaturesHelper._build_sequence_encoder(self.dataset, p_value, encoder_params)

        with encoder.relevant_indices_path.open("rb") as file:
            relevant_indices = pickle.load(file)

        relevant_feature_presence = self._sum_sequence_vectors_iterable(relevant_indices, encoder.comparison_data.get_item_vectors())

        return self._get_positive_negative_class(relevant_feature_presence, self.dataset.get_repertoire_ids())

    def _compute_significant_compairr_sequence_counts(self, p_value, encoder_params):
        encoder = SignificantFeaturesHelper._build_compairr_sequence_encoder(self.dataset, p_value, encoder_params, self.compairr_path)

        with encoder.relevant_indices_path.open("rb") as file:
            relevant_indices = pickle.load(file)

        relevant_feature_presence = self._sum_sequence_vectors_iterable(relevant_indices, encoder.compairr_sequence_presence)

        return self._get_positive_negative_class(relevant_feature_presence, self.dataset.get_repertoire_ids())

    def _sum_sequence_vectors_iterable(self, relevant_indices, sequence_vector_iterable):
        relevant_feature_presence = np.zeros(shape=(6,))

        for i, sequence_vector in enumerate(sequence_vector_iterable):
            if relevant_indices[i]:
                relevant_feature_presence += sequence_vector

        return relevant_feature_presence

    def _get_relevant_feature_presence(self, encoder, relevant_indices):

        if isinstance(encoder, KmerAbundanceEncoder):
            relevant_feature_presence = np.sum(encoder.kmer_presence_matrix[relevant_indices], axis=0)
        elif isinstance(encoder, CompAIRRSequenceAbundanceEncoder):
            relevant_feature_presence = np.sum(encoder.sequence_presence_matrix[relevant_indices], axis=0)
        else:
            relevant_feature_presence = np.zeros(shape=(6,))

            for i, sequence_vector in enumerate(encoder.comparison_data.get_item_vectors()):
                if relevant_indices[i]:
                    relevant_feature_presence += sequence_vector

        return relevant_feature_presence

    def _get_positive_negative_class(self, relevant_feature_presence, repertoire_ids):
        is_positive_class = AbundanceEncoderHelper.check_is_positive_class(self.dataset, repertoire_ids, self.label_config)

        pos_class_feature_counts = relevant_feature_presence[is_positive_class]
        neg_class_feature_counts = relevant_feature_presence[np.logical_not(is_positive_class)]

        return pos_class_feature_counts, neg_class_feature_counts


import numpy as np

from immuneML.data_model.dataset.ReceptorDataset import ReceptorDataset
from immuneML.data_model.encoded_data.EncodedData import EncodedData
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.onehot.OneHotEncoder import OneHotEncoder
from immuneML.util.EncoderHelper import EncoderHelper


class OneHotReceptorEncoder(OneHotEncoder):
    """
    One-hot encoded repertoire data is represented in a matrix with dimensions:
        [receptors, chains, sequence_lengths, one_hot_characters]

    when use_positional_info is true, the last 3 indices in one_hot_characters represents the positional information:
        - start position (high when close to start)
        - middle position (high in the middle of the sequence)
        - end position (high when close to end)
    """

    def _encode_new_dataset(self, dataset, params: EncoderParams):
        encoded_data = self._encode_data(dataset, params)

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = encoded_data

        return encoded_dataset

    def _encode_data(self, dataset: ReceptorDataset, params: EncoderParams):
        receptor_objs = [receptor for receptor in dataset.get_data()]
        sequences = [[getattr(obj, chain).get_sequence(self.sequence_type) for chain in obj.get_chains()] for obj in receptor_objs]
        first_chain_seqs, second_chain_seqs = zip(*sequences)

        if any(seq is None for seq in first_chain_seqs) or any(seq is None for seq in second_chain_seqs):
            raise ValueError(f"{OneHotEncoder.__name__}: receptor dataset {dataset.name} (id: {dataset.identifier}) contains empty sequences for the "
                             f"specified sequence type {self.sequence_type.name.lower()}. Please check that the dataset is imported correctly.")

        max_seq_len = max(max([len(seq) for seq in first_chain_seqs]), max([len(seq) for seq in second_chain_seqs]))

        labels = self._get_labels(receptor_objs, params) if params.encode_labels else None

        examples_first_chain = self._encode_sequence_list(first_chain_seqs, pad_n_sequences=len(receptor_objs),
                                                          pad_sequence_len=max_seq_len)
        examples_second_chain = self._encode_sequence_list(second_chain_seqs, pad_n_sequences=len(receptor_objs),
                                                           pad_sequence_len=max_seq_len)

        examples = np.stack((examples_first_chain, examples_second_chain), axis=1)

        feature_names = self._get_feature_names(max_seq_len, receptor_objs[0].get_chains())

        if self.flatten:
            examples = examples.reshape((len(receptor_objs), 2*max_seq_len*len(self.onehot_dimensions)))
            feature_names = [item for sublist in feature_names for subsublist in sublist for item in subsublist]

        encoded_data = EncodedData(examples=examples,
                                   labels=labels,
                                   example_ids=dataset.get_example_ids(),
                                   feature_names=feature_names,
                                   encoding=OneHotEncoder.__name__,
                                   example_weights=dataset.get_example_weights(),
                                   info={"chain_names": receptor_objs[0].get_chains() if all(receptor_obj.get_chains() == receptor_objs[0].get_chains() for receptor_obj in receptor_objs) else None})

        return encoded_data

    def _get_feature_names(self, max_seq_len, chains):
        return [[[f"{chain}_{pos}_{dim}" for dim in self.onehot_dimensions] for pos in range(max_seq_len)] for chain in chains]

    def _get_labels(self, receptor_objs, params: EncoderParams):
        label_names = params.label_config.get_labels_by_name()
        labels = {name: [None] * len(receptor_objs) for name in label_names}

        for idx, receptor in enumerate(receptor_objs):
            for name in label_names:
                labels[name][idx] = receptor.metadata[name]

        return labels



from importlib import import_module, util
from pathlib import Path

from immuneML.environment.EnvironmentSettings import EnvironmentSettings


class ReflectionHandler:

    @staticmethod
    def import_function(function: str, module):
        return getattr(module, function)

    @staticmethod
    def import_module(name: str, package: str = None):
        return import_module(name, package)

    @staticmethod
    def get_class_from_path(path, class_name: str = None):
        """ obtain the class reference from the given path

        Args:

            path (str or pathlib.Path): path to file where the class is located
            class_name (str): class name to import_dataset from the file; if None, it is assumed that the class name is the same as the file name

        Returns:
             class
        """
        path = Path(path)
        if class_name is None:
            class_name = path.stem

        return ReflectionHandler._import_class(path, class_name)

    @staticmethod
    def _import_class(path: Path, class_name: str):
        module_path = ".".join(path.parts[len(list(path.parts)) - list(path.parts)[::-1].index("immuneML") - 1:])[:-3]
        mod = import_module(module_path)
        cls = getattr(mod, class_name)
        return cls

    @staticmethod
    def get_class_by_name(class_name: str, subdirectory: str = ""):
        filenames = ReflectionHandler._get_filenames(class_name, subdirectory)

        assert len(filenames) == 1, f"ReflectionHandler could not find class named {class_name}. Check spelling and try again."

        return ReflectionHandler._import_class(filenames[0], class_name)

    @staticmethod
    def _get_filenames(class_name: str, subdirectory_name: str = "", partial=False):
        pattern = f"immuneML/**/*{class_name}.py" if partial else f"immuneML/**/{class_name}.py"

        filenames = list(EnvironmentSettings.root_path.glob(pattern))
        filenames = [f for f in filenames if subdirectory_name in "/".join(f.parts)]

        return filenames

    @staticmethod
    def all_subclasses(cls):
        subclasses = set(cls.__subclasses__()).union([s for c in cls.__subclasses__() for s in ReflectionHandler.all_subclasses(c)])
        return subclasses

    @staticmethod
    def all_direct_subclasses(cls, drop_part=None, subdirectory=None):
        if drop_part is not None and subdirectory is not None:
            classes = ReflectionHandler.get_classes_by_partial_name(drop_part, subdirectory)
        return [cl for cl in cls.__subclasses__()]

    @staticmethod
    def all_nonabstract_subclass_basic_names(cls, drop_part: str, subdirectory: str = ""):
        return [c.__name__.replace(drop_part, "") for c in ReflectionHandler.all_nonabstract_subclasses(cls, drop_part, subdirectory)]

    @staticmethod
    def all_nonabstract_subclasses(cls, drop_part=None, subdirectory=None):
        if drop_part is not None and subdirectory is not None:
            classes = ReflectionHandler.get_classes_by_partial_name(drop_part, subdirectory)
        return [cl for cl in ReflectionHandler.all_subclasses(cls) if not bool(getattr(cl, "__abstractmethods__", False))]

    @staticmethod
    def exists(class_name: str, subdirectory: str = ""):
        filenames = ReflectionHandler._get_filenames(class_name, subdirectory)

        if len(filenames) == 1:
            return True
        else:
            return False

    @staticmethod
    def discover_classes_by_partial_name(class_name_ending: str, subdirectory: str = ""):
        filenames = ReflectionHandler._get_filenames(class_name_ending, subdirectory, partial=True)
        class_names = [f.stem for f in filenames]

        return class_names

    @staticmethod
    def get_classes_by_partial_name(class_name_ending: str, subdirectory: str = ""):
        filenames = ReflectionHandler._get_filenames(class_name_ending, subdirectory, partial=True)
        classes = [ReflectionHandler._import_class(filename, filename.stem) for filename in filenames]
        return classes

    @staticmethod
    def is_installed(module_name: str) -> bool:
        return bool(util.find_spec(module_name))



import copy
import logging
import warnings
from multiprocessing.pool import Pool
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd

from immuneML.IO.dataset_export.ImmuneMLExporter import ImmuneMLExporter
from immuneML.IO.dataset_import.DatasetImportParams import DatasetImportParams
from immuneML.IO.dataset_import.ImmuneMLImport import ImmuneMLImport
from immuneML.data_model.dataset import Dataset
from immuneML.data_model.dataset.ReceptorDataset import ReceptorDataset
from immuneML.data_model.dataset.RepertoireDataset import RepertoireDataset
from immuneML.data_model.dataset.SequenceDataset import SequenceDataset
from immuneML.data_model.receptor.BCKReceptor import BCKReceptor
from immuneML.data_model.receptor.BCReceptor import BCReceptor
from immuneML.data_model.receptor.ChainPair import ChainPair
from immuneML.data_model.receptor.Receptor import Receptor
from immuneML.data_model.receptor.RegionType import RegionType
from immuneML.data_model.receptor.TCABReceptor import TCABReceptor
from immuneML.data_model.receptor.TCGDReceptor import TCGDReceptor
from immuneML.data_model.receptor.receptor_sequence.Chain import Chain
from immuneML.data_model.receptor.receptor_sequence.ReceptorSequence import ReceptorSequence
from immuneML.data_model.receptor.receptor_sequence.SequenceFrameType import SequenceFrameType
from immuneML.data_model.receptor.receptor_sequence.SequenceMetadata import SequenceMetadata
from immuneML.data_model.repertoire.Repertoire import Repertoire
from immuneML.environment.Constants import Constants
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.environment.SequenceType import SequenceType
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class ImportHelper:
    DATASET_FORMAT = "yaml"

    @staticmethod
    def import_dataset(import_class, params: dict, dataset_name: str) -> Dataset:
        processed_params = DatasetImportParams.build_object(**params)

        dataset = ImportHelper.load_dataset_if_exists(params, processed_params, dataset_name)
        if dataset is None:
            # backwards compatibility: if is_repertoire is not specified but the metadata file is
            if processed_params.is_repertoire is None and processed_params.metadata_file is not None:
                processed_params.is_repertoire = True

            if processed_params.is_repertoire:
                dataset = ImportHelper.import_repertoire_dataset(import_class, processed_params, dataset_name)
            else:
                dataset = ImportHelper.import_sequence_dataset(import_class, processed_params, dataset_name)

        return dataset

    @staticmethod
    def load_dataset_if_exists(params: dict, processed_params, dataset_name: str):

        dataset_file = processed_params.result_path / f"{dataset_name}.{ImportHelper.DATASET_FORMAT}"
        dataset = None

        if dataset_file.is_file():
            params["path"] = dataset_file
            dataset = ImmuneMLImport.import_dataset(params, dataset_name)

        return dataset

    @staticmethod
    def import_repertoire_dataset(import_class, params: DatasetImportParams, dataset_name: str) -> RepertoireDataset:
        """
        Function to create a dataset from the metadata and a list of repertoire files and exports dataset pickle file

        Arguments:
            import_class: class to use for import
            params: instance of DatasetImportParams class which includes information on path, columns, result path etc.
            dataset_name: user-defined name of the dataset

        Returns:
            RepertoireDataset object that was created

        """

        try:
            metadata = pd.read_csv(params.metadata_file, sep=",")
        except Exception as e:
            raise Exception(f"{e}\nAn error occurred while reading in the metadata file {params.metadata_file}. Please "
                            f"see the error log above for more details on this error and the documentation for the "
                            f"expected format of the metadata.")

        ParameterValidator.assert_keys_present(metadata.columns.tolist(), ["filename"], ImportHelper.__name__,
                                               f'{dataset_name}: params: metadata_file')

        PathBuilder.build(params.result_path / "repertoires/")

        arguments = [(import_class, row, params) for index, row in metadata.iterrows()]
        with Pool(params.number_of_processes) as pool:
            repertoires = pool.starmap(ImportHelper.load_repertoire_as_object, arguments)

        new_metadata_file = ImportHelper.make_new_metadata_file(repertoires, metadata, params.result_path, dataset_name)

        potential_labels = list(set(metadata.columns.tolist()) - {"filename"})
        dataset = RepertoireDataset(labels={key: list(set(metadata[key].values.tolist())) for key in potential_labels},
                                    repertoires=repertoires, metadata_file=new_metadata_file, name=dataset_name)

        ImmuneMLExporter.export(dataset, params.result_path)

        return dataset

    @staticmethod
    def make_new_metadata_file(repertoires: list, metadata: pd.DataFrame, result_path: Path, dataset_name: str) -> Path:
        new_metadata = metadata.copy()
        new_metadata.loc[:, "filename"] = [repertoire.data_filename.name for repertoire in repertoires]
        new_metadata.loc[:, "identifier"] = [repertoire.identifier for repertoire in repertoires]

        metadata_filename = result_path / f"{dataset_name}_metadata.csv"
        new_metadata.to_csv(metadata_filename, index=False, sep=",")

        return metadata_filename

    @staticmethod
    def load_repertoire_as_object(import_class, metadata_row, params: DatasetImportParams):
        try:
            alternative_load_func = getattr(import_class, "alternative_load_func", None)

            filename = params.path / f"{metadata_row['filename']}"
            if not filename.is_file():
                filename = params.path / f"repertoires/{metadata_row['filename']}"

            dataframe = ImportHelper.load_sequence_dataframe(filename, params, alternative_load_func)
            dataframe = import_class.preprocess_dataframe(dataframe, params)
            sequence_lists = {field: dataframe[field].values.tolist() for field in Repertoire.FIELDS if
                              field in dataframe.columns}
            sequence_lists = {**sequence_lists, **{field: dataframe[field].values.tolist()
                                                   for field in list(set(dataframe.columns) - set(Repertoire.FIELDS))}}

            repertoire_inputs = {**{"metadata": metadata_row.to_dict(),
                                    "path": params.result_path / "repertoires/",
                                    "filename_base": filename.stem}, **sequence_lists}
            repertoire = Repertoire.build(**repertoire_inputs)

            return repertoire
        except Exception as exception:
            raise RuntimeError(
                f"{ImportHelper.__name__}: error when importing file {metadata_row['filename']}: {exception}") from exception

    @staticmethod
    def load_sequence_dataframe(filepath, params, alternative_load_func=None):
        try:
            if alternative_load_func:
                df = alternative_load_func(filepath, params)
            else:
                df = ImportHelper.safe_load_dataframe(filepath, params)
        except Exception as ex:
            raise Exception(
                f"{ex}\n\nImportHelper: an error occurred during dataset import while parsing the input file: {filepath}.\n"
                f"Please make sure this is a correct immune receptor data file (not metadata).\n"
                f"The parameters used for import are {params}.\nFor technical description of the error, see the log above. "
                f"For details on how to specify the dataset import, see the documentation.")

        ImportHelper.rename_dataframe_columns(df, params)
        ImportHelper.standardize_none_values(df)

        return df

    @staticmethod
    def safe_load_dataframe(filepath, params: DatasetImportParams):
        if hasattr(params, "columns_to_load") and params.columns_to_load is not None:
            usecols = set(params.columns_to_load) if hasattr(params,
                                                             "columns_to_load") and params.columns_to_load is not None else set()
            usecols = usecols.union(
                set(params.column_mapping.keys()) if hasattr(params,
                                                             "column_mapping") and params.column_mapping is not None else set())
            usecols = usecols.union(set(params.column_mapping_synonyms.keys())
                                    if hasattr(params,
                                               "column_mapping_synonyms") and params.column_mapping_synonyms is not None else set())
            usecols = usecols.union(set(params.metadata_column_mapping.keys())
                                    if hasattr(params,
                                               "metadata_column_mapping") and params.metadata_column_mapping is not None else set())
        else:
            usecols = None

        try:
            df = pd.read_csv(filepath, sep=params.separator, iterator=False, usecols=usecols, dtype=str)
        except ValueError:
            try:
                df = pd.read_csv(filepath, sep=params.separator, iterator=False, usecols=params.columns_to_load, dtype=str)
            except ValueError:
                df = pd.read_csv(filepath, sep=params.separator, iterator=False, dtype=str)

                expected = [e for e in params.columns_to_load if e not in list(df.columns)]

                warnings.warn(
                    f"ImportHelper: expected to find the following column(s) in the input file '{filepath.name}', which were not found: {expected}."
                    f"The following columns were imported instead: {list(df.columns)}. \nTo remove this warning, add the relevant columns "
                    f"to the input file, or change which columns are imported under 'datasets/<dataset_key>/params/columns_to_load' and 'datasets/<dataset_key>/params/column_mapping'.")

        return df

    @staticmethod
    def rename_dataframe_columns(df, params: DatasetImportParams):
        if hasattr(params, "column_mapping") and params.column_mapping is not None:
            df.rename(columns=params.column_mapping, inplace=True)

        if hasattr(params, "column_mapping_synonyms") and params.column_mapping_synonyms is not None:
            for synonym, colname in params.column_mapping_synonyms.items():
                if colname not in df.columns:
                    df.rename(columns={synonym: colname}, inplace=True)

        if hasattr(params, "metadata_column_mapping") and params.metadata_column_mapping is not None:
            df.rename(columns=params.metadata_column_mapping, inplace=True)

        invalid_chars = [" ", "#", "&"]
        invalid_col_names = {col: col.replace(" ", "_").replace("#", "_").replace("&", "_")
                             for col in df.columns if any(el in col for el in invalid_chars)}
        if len(invalid_col_names.keys()) > 0:
            logging.warning(
                f"Note that column names that contain characters which are not letters, numbers nor '_' signs"
                f" have been renamed to replace these invalid characters with '_' instead: {invalid_col_names}")
        df.rename(columns=invalid_col_names, inplace=True)

    @staticmethod
    def standardize_none_values(dataframe: pd.DataFrame):
        dataframe.replace(
            {key: Constants.UNKNOWN for key in ["unresolved", "no data", "na", "unknown", "null", "nan", np.nan, ""]},
            inplace=True)

    @staticmethod
    def drop_empty_sequences(dataframe: pd.DataFrame, import_empty_aa_sequences: bool,
                             import_empty_nt_sequences: bool) -> pd.DataFrame:
        sequence_types = []
        if not import_empty_aa_sequences:
            sequence_types.append(SequenceType.AMINO_ACID)
        if not import_empty_nt_sequences:
            sequence_types.append(SequenceType.NUCLEOTIDE)

        for sequence_type in sequence_types:
            sequence_colname = sequence_type.value
            sequence_name = sequence_type.name.lower().replace("_", " ")

            if sequence_colname in dataframe.columns:
                dataframe[sequence_colname].replace({"": Constants.UNKNOWN}, inplace=True)

                n_empty = sum(dataframe[sequence_colname].isnull())
                if n_empty > 0:
                    dataframe.drop(dataframe.loc[dataframe[sequence_colname].isnull()].index, inplace=True)
                    warnings.warn(
                        f"{ImportHelper.__name__}: {n_empty} sequences were removed from the dataset because they contained an empty {sequence_name} "
                        f"sequence after preprocessing. ")
            else:
                warnings.warn(
                    f"{ImportHelper.__name__}: column {sequence_colname} was not set, but is required for filtering. Skipping this filtering...")

        return dataframe

    @staticmethod
    def drop_illegal_character_sequences(dataframe: pd.DataFrame, import_illegal_characters: bool,
                                         import_with_stop_codon: bool) -> pd.DataFrame:
        for sequence_type in SequenceType:
            if not import_illegal_characters:
                sequence_name = sequence_type.name.lower().replace("_", " ")

                legal_alphabet = EnvironmentSettings.get_sequence_alphabet(sequence_type)
                if sequence_type == SequenceType.AMINO_ACID and import_with_stop_codon:
                    legal_alphabet.append(Constants.STOP_CODON)

                if sequence_type.value in dataframe.columns:
                    is_illegal_seq = [ImportHelper.is_illegal_sequence(sequence, legal_alphabet) for sequence in
                                      dataframe[sequence_type.value]]
                    n_illegal = sum(is_illegal_seq)
                    n_total = dataframe.shape[0]

                    if n_illegal > 0:
                        dataframe.drop(dataframe.loc[is_illegal_seq].index, inplace=True)
                        warnings.warn(
                            f"{ImportHelper.__name__}: {n_illegal}/{n_total} sequences were removed from the dataset because their {sequence_name}"
                            f" sequence contained illegal characters. ")

                else:
                    logging.warning(
                        f"{ImportHelper.__name__}: column {sequence_type.value} is missing, illegal characters were not checked.")

        return dataframe

    @staticmethod
    def is_illegal_sequence(sequence, legal_alphabet) -> bool:
        if sequence is None:
            return False
        else:
            return not all(character in legal_alphabet for character in sequence)

    @staticmethod
    def prepare_frame_type_list(params: DatasetImportParams) -> list:
        frame_type_list = []
        if params.import_productive:
            frame_type_list.append(SequenceFrameType.IN.value)
        if params.import_unknown_productivity:
            frame_type_list.append(SequenceFrameType.UNDEFINED.value)
        if params.import_out_of_frame:
            frame_type_list.append(SequenceFrameType.OUT.value)
        if params.import_with_stop_codon:
            frame_type_list.append(SequenceFrameType.STOP.value)
        return frame_type_list

    @staticmethod
    def load_chains(df: pd.DataFrame):
        if "chain" in df.columns:
            df["chain"] = ImportHelper.load_chains_from_chains(df)
        else:
            df["chain"] = ImportHelper.load_chains_from_genes(df)

    @staticmethod
    def load_chains_from_chains(df: pd.DataFrame) -> list:
        return [Chain.get_chain(chain_str).value if chain_str is not None else None for chain_str in df["chain"]]

    @staticmethod
    def load_chains_from_genes(df: pd.DataFrame) -> list:
        return df.apply(ImportHelper.get_chain_for_row, axis=1)

    @staticmethod
    def get_chain_for_row(row):
        for col in ["v_call", "j_call"]:
            if col in row and row[col] is not None:
                chain = Chain.get_chain(str(row[col])[0:3])
                return chain.value if chain is not None else None
        return None

    @staticmethod
    def junction_to_cdr3(df: pd.DataFrame, region_type: RegionType):
        """
        If RegionType is CDR3, the leading C and trailing W are removed from the sequence to match the IMGT CDR3 definition.
        This method alters the data in the provided dataframe.
        """

        if region_type == RegionType.IMGT_CDR3:
            if "sequence_aa" in df:
                df.loc[:, "sequence_aa"] = df["sequence_aa"].str[1:-1]
            if "sequence" in df:
                df.loc[:, "sequence"] = df["sequence"].str[3:-3]
            df["region_type"] = region_type.name

    @staticmethod
    def strip_alleles(df: pd.DataFrame, column_name):
        return ImportHelper.strip_suffix(df, column_name, Constants.ALLELE_DELIMITER)

    @staticmethod
    def strip_genes(df: pd.DataFrame, column_name):
        return ImportHelper.strip_suffix(df, column_name, Constants.GENE_DELIMITER)

    @staticmethod
    def strip_suffix(df: pd.DataFrame, column_name, delimiter):
        """
        Safely removes everything after a delimiter from a column in the DataFrame
        """
        if column_name in df.columns:
            return df[column_name].apply(lambda gene_col: None if gene_col is None else gene_col.rsplit(delimiter)[0])

    @staticmethod
    def get_sequence_filenames(path: Path, dataset_name: str):
        data_file_extensions = ("*.tsv", "*.csv", "*.txt")

        if path.is_file():
            filenames = [path]
        elif path.is_dir():
            filenames = []

            for pattern in data_file_extensions:
                filenames.extend(list(path.glob(pattern)))
        else:
            raise ValueError(f"ImportHelper: path '{path}' given in YAML specification is not a valid path. "
                             f"This parameter can either point to a single file with immune receptor data or to a directory containing such files.")

        assert len(
            filenames) >= 1, f"ImportHelper: the dataset {dataset_name} cannot be imported, no files were found under {path}.\n" \
                             f"Note that only files with the following extensions can be imported: {data_file_extensions}"
        return filenames

    @staticmethod
    def import_sequence_dataset(import_class, params, dataset_name: str):
        PathBuilder.build(params.result_path)

        filenames = ImportHelper.get_sequence_filenames(params.path, dataset_name)

        dataset_params = {}
        items = None

        for index, filename in enumerate(filenames):
            new_items = ImportHelper.import_items(import_class, filename, params)
            items = np.append(items, new_items) if items is not None else new_items
            dataset_params = ImportHelper.extract_sequence_dataset_params(items, params)

        cls = ReceptorDataset if params.paired else SequenceDataset
        dataset = cls.build_from_objects(items, params.sequence_file_size, params.result_path, dataset_name,
                                         dataset_params)

        ImmuneMLExporter.export(dataset, params.result_path)

        return dataset

    @staticmethod
    def extract_sequence_dataset_params(items=None, params=None) -> dict:
        result = {}
        if params is not None:
            result = {'region_type': params.region_type, 'receptor_chains': params.receptor_chains,
                      'organism': params.organism}
        if items is not None:
            for index, item in enumerate(items):
                metadata = item.metadata if params.paired else item.metadata.custom_params if item.metadata is not None else {}
                for key in metadata:
                    if key in result and isinstance(result[key], set):
                        result[key].add(metadata[key])
                    elif key not in result:
                        result[key] = {metadata[key]}
        return result

    @staticmethod
    def import_items(import_class, path, params: DatasetImportParams):
        alternative_load_func = getattr(import_class, "alternative_load_func", None)
        df = ImportHelper.load_sequence_dataframe(path, params, alternative_load_func)
        df = import_class.preprocess_dataframe(df, params)

        if params.paired:
            import_receptor_func = getattr(import_class, "import_receptors", None)
            if import_receptor_func:
                sequences = import_receptor_func(df, params)
            else:
                raise NotImplementedError(
                    f"{import_class.__name__}: import of paired receptor data has not been implemented.")
        else:
            metadata_columns = params.metadata_column_mapping.values() if params.metadata_column_mapping else None
            sequences = df.apply(ImportHelper.import_sequence, metadata_columns=metadata_columns, axis=1).values

        return sequences

    @staticmethod
    def import_sequence(row, metadata_columns=None) -> ReceptorSequence:
        if metadata_columns is None:
            metadata_columns = []
        metadata = SequenceMetadata(
            v_call=str(row["v_call"]) if "v_call" in row and row["v_call"] is not None else None,
            j_call=str(row["j_call"]) if "j_call" in row and row["j_call"] is not None else None,
            cell_id=str(row['cell_id']) if 'cell_id' in row and row['cell_id'] is not None else None,
            chain=row["chain"] if "chain" in row and row["chain"] is not None else None,
            region_type=row["region_type"] if "region_type" in row and row["region_type"] is not None else None,
            duplicate_count=int(row["duplicate_count"]) if "duplicate_count" in row and row[
                "duplicate_count"] is not None else None,
            frame_type=row["frame_type"] if "frame_type" in row and row["frame_type"] is not None else None,
            custom_params={custom_col: row[custom_col] for custom_col in metadata_columns if
                           custom_col in row} if metadata_columns is not None else {})
        sequence = ReceptorSequence(
            sequence_aa=str(row["sequence_aa"]) if "sequence_aa" in row and row[
                "sequence_aa"] is not None else None,
            sequence=str(row["sequence"]) if "sequence" in row and row["sequence"] is not None else None,
            sequence_id=str(row["sequence_id"]) if "sequence_id" in row and row["sequence_id"] is not None else None,
            metadata=metadata)

        return sequence

    @staticmethod
    def import_receptors(df, params) -> List[Receptor]:
        identifiers = df["receptor_id"].unique()

        chain_pair = params.receptor_chains
        if chain_pair is None:
            chains = [Chain.get_chain(chain) for chain in df["chain"].unique()]
            chain_pair = ChainPair.get_chain_pair(chains)

        metadata_columns = list(params.metadata_column_mapping.values()) if params.metadata_column_mapping else None

        all_receptors = []

        for identifier in identifiers:
            receptors = ImportHelper.import_receptors_by_id(df, identifier, chain_pair, metadata_columns)
            all_receptors.extend(receptors)

        return all_receptors

    @staticmethod
    def import_receptors_by_id(df, identifier, chain_pair, metadata_columns) -> List[Receptor]:
        first_row = df.loc[(df["receptor_id"] == identifier) & (df["chain"] == chain_pair.value[0])]
        second_row = df.loc[(df["receptor_id"] == identifier) & (df["chain"] == chain_pair.value[1])]

        for i, row in enumerate([first_row, second_row]):
            if row.shape[0] > 1:
                warnings.warn(
                    f"Multiple {chain_pair.value[i]} chains found for receptor with identifier {identifier}, only the first entry will be loaded")
            elif row.shape[0] == 0:
                warnings.warn(
                    f"Missing {chain_pair.value[i]} chain for receptor with identifier {identifier}, this receptor will be omitted.")
                return []

        # todo add possibility to import multiple chain combo's? (BCR heavy-light & heavy-kappa, as seen in 10xGenomics?)

        return [ImportHelper.build_receptor_from_rows(first_row.iloc[0], second_row.iloc[0], identifier, chain_pair,
                                                      metadata_columns)]

    @staticmethod
    def build_receptor_from_rows(first_row, second_row, identifier, chain_pair, metadata_columns):
        first_sequence = ImportHelper.import_sequence(first_row, metadata_columns=metadata_columns)
        second_sequence = ImportHelper.import_sequence(second_row, metadata_columns=metadata_columns)

        if chain_pair == ChainPair.TRA_TRB:
            receptor = TCABReceptor(alpha=first_sequence,
                                    beta=second_sequence,
                                    identifier=identifier,
                                    metadata=copy.deepcopy(second_sequence.metadata.custom_params))
        elif chain_pair == ChainPair.TRG_TRD:
            receptor = TCGDReceptor(gamma=first_sequence,
                                    delta=second_sequence,
                                    identifier=identifier,
                                    metadata=copy.deepcopy(second_sequence.metadata.custom_params))
        elif chain_pair == ChainPair.IGH_IGL:
            receptor = BCReceptor(heavy=first_sequence,
                                  light=second_sequence,
                                  identifier=identifier,
                                  metadata=copy.deepcopy(first_sequence.metadata.custom_params))
        elif chain_pair == ChainPair.IGH_IGK:
            receptor = BCKReceptor(heavy=first_sequence,
                                   kappa=second_sequence,
                                   identifier=identifier,
                                   metadata=copy.deepcopy(first_sequence.metadata.custom_params))
        else:
            raise NotImplementedError(f"ImportHelper: {chain_pair} chain pair is not supported.")

        return receptor



# quality: gold
import logging
import shutil
import weakref
from pathlib import Path
from typing import List, Any, Dict
from uuid import uuid4

import bionumpy as bnp
import numpy as np
import yaml

from immuneML.data_model.DatasetItem import DatasetItem
from immuneML.data_model.SequenceSet import SequenceSet
from immuneML.data_model.bnp_util import bnp_write_to_file, write_yaml, bnp_read_from_file, \
    make_dynamic_seq_set_dataclass, build_dynamic_bnp_dataclass_obj
from immuneML.data_model.receptor.ChainPair import ChainPair
from immuneML.data_model.receptor.Receptor import Receptor
from immuneML.data_model.receptor.ReceptorBuilder import ReceptorBuilder
from immuneML.data_model.receptor.RegionType import RegionType
from immuneML.data_model.receptor.receptor_sequence.Chain import Chain
from immuneML.data_model.receptor.receptor_sequence.ReceptorSequence import ReceptorSequence
from immuneML.data_model.receptor.receptor_sequence.SequenceMetadata import SequenceMetadata
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.util.NumpyHelper import NumpyHelper
from immuneML.util.PathBuilder import PathBuilder


class Repertoire(DatasetItem):
    """
    Repertoire object consisting of sequence objects, each sequence attribute is stored as a list across all sequences and can be
    loaded separately. Internally, this class relies on numpy to store/import_dataset the data.
    """

    FIELDS = ("sequence_aa", "sequence", "v_call", "j_call", "chain", "duplicate_count", "region_type", "frame_type",
              "sequence_id", "cell_id")

    @classmethod
    def build(cls, path: Path = None, metadata: dict = None, filename_base: str = None, identifier: str = None, **kwargs):

        identifier = uuid4().hex if identifier is None else identifier
        filename_base = filename_base if filename_base is not None else identifier
        data_filename = path / f"{filename_base}.tsv"

        bnp_object, type_dict = build_dynamic_bnp_dataclass_obj(kwargs)
        bnp_write_to_file(data_filename, bnp_object)

        metadata_filename = path / f"{filename_base}_metadata.yaml"
        metadata = {} if metadata is None else metadata
        metadata['type_dict'] = {key: SequenceSet.TYPE_TO_STR[val] for key, val in type_dict.items()}
        write_yaml(metadata_filename, metadata)

        repertoire = Repertoire(data_filename, metadata_filename, identifier)
        return repertoire

    @classmethod
    def build_like(cls, repertoire: 'Repertoire', indices_to_keep: list, result_path: Path, filename_base: str = None):
        PathBuilder.build(result_path)

        identifier = uuid4().hex
        filename_base = filename_base if filename_base is not None else identifier

        data_filename = result_path / f"{filename_base}.tsv"
        bnp_data = repertoire.load_bnp_data()
        bnp_data = bnp_data[indices_to_keep]

        bnp_write_to_file(data_filename, bnp_data)

        metadata_filename = result_path / f"{filename_base}_metadata.yaml"
        shutil.copyfile(repertoire.metadata_filename, metadata_filename)

        new_repertoire = Repertoire(data_filename, metadata_filename, identifier)
        return new_repertoire

    @classmethod
    def build_from_sequence_objects(cls, sequence_objects: list, path: Path, metadata: dict, filename_base: str = None,
                                    repertoire_id: str = None):

        assert all(isinstance(sequence, ReceptorSequence) for sequence in sequence_objects), \
            "Repertoire: all sequences have to be instances of ReceptorSequence class."

        sequence_aa, sequence, v_call, j_call, chain, duplicate_count, region_type, frame_type, sequence_id, cell_id = [], [], [], [], [], [], [], [], [], []
        custom_lists = {key: [] for key in sequence_objects[0].metadata.custom_params} if sequence_objects[
            0].metadata else {}
        signals = {}

        for index, seq in enumerate(sequence_objects):
            sequence_id.append(seq.sequence_id)
            sequence_aa.append(seq.sequence_aa)
            sequence.append(seq.sequence)
            if seq.metadata:
                v_call.append(seq.metadata.v_call)
                j_call.append(seq.metadata.j_call)
                chain.append(seq.metadata.chain.value if seq.metadata.chain else None)
                duplicate_count.append(seq.metadata.duplicate_count)
                region_type.append(seq.metadata.region_type.value if seq.metadata.region_type else None)
                frame_type.append(seq.metadata.frame_type.value if seq.metadata.frame_type else None)
                cell_id.append(seq.metadata.cell_id)
                for param in seq.metadata.custom_params.keys():
                    current_value = seq.metadata.custom_params[param] if param in seq.metadata.custom_params else None
                    if param in custom_lists:
                        custom_lists[param].append(current_value)
                    else:
                        custom_lists[param] = [None for _ in range(index)]
                        custom_lists[param].append(current_value)

        sequence_count = len(sequence)

        for signal in signals.keys():
            signal_info_count = len(signals[signal])
            if signal_info_count < sequence_count:
                signals[signal].extend([None for _ in range(sequence_count - signal_info_count)])

        return cls.build(sequence_aa=sequence_aa, sequence=sequence, v_call=v_call, j_call=j_call, chain=chain,
                         duplicate_count=duplicate_count, region_type=region_type, frame_type=frame_type,
                         sequence_id=sequence_id, path=path, metadata=metadata, cell_id=cell_id,
                         filename_base=filename_base, identifier=repertoire_id, **custom_lists, **signals)

    @property
    def _bnp_filename(self):
        return Path(str(self.data_filename) + ".tsv")

    def __init__(self, data_filename: Path, metadata_filename: Path, identifier: str):
        data_filename = Path(data_filename)
        metadata_filename = Path(metadata_filename) if metadata_filename is not None else None

        assert data_filename.suffix == ".tsv", \
            f"Repertoire: the file representing the repertoire has to be in tsv format. Got {data_filename.suffix} instead."

        self.data_filename = data_filename
        if metadata_filename:
            with metadata_filename.open("r") as file:
                self.metadata = yaml.safe_load(file)

        self.metadata_filename = metadata_filename
        self.identifier = identifier
        self.bnp_data = None
        self.element_count = None

    @property
    def _type_dict(self):
        return {key: SequenceSet.STR_TO_TYPE[val] for key, val in self.metadata["type_dict"].items()}

    @property
    def _buffer_type(self):
        return self._create_buffer_type_from_field_dict(self._type_dict)

    def get_sequence_aas(self):
        return self.get_attribute("sequence_aa")

    def get_sequence_identifiers(self):
        return self.get_attribute("sequence_id")

    def get_v_genes(self):
        return self.get_attribute("v_call")

    def get_j_genes(self):
        return self.get_attribute("j_call")

    def get_counts(self):
        counts = self.get_attribute("duplicate_count")
        if counts is not None:
            counts = np.array([int(count) if count != SequenceSet.get_neutral_value(int) else None for count in counts])
        return counts

    def get_chains(self):
        chains = self.get_attribute("chain")
        if chains is not None:
            chains = np.array(
                [Chain.get_chain(chain_str) if chain_str is not None else None for chain_str in chains.tolist()])
        return chains

    def get_attribute(self, attribute, as_list: bool = False):
        data = self.load_bnp_data()
        if attribute in self._type_dict:
            tmp = getattr(data, attribute)
            if as_list:
                return tmp.tolist()
            else:
                return tmp
        else:
            return None

    def get_attributes(self, attributes: list, as_list: bool = False):
        result = {}
        data = self.load_bnp_data()
        for attribute in attributes:
            if attribute in self._type_dict:
                result[attribute] = getattr(data, attribute).tolist() if as_list else getattr(data, attribute)
            else:
                logging.warning(
                    f"{Repertoire.__name__}: attribute {attribute} is not present in the repertoire {self.identifier}, skipping...")
        return result

    def get_region_type(self):
        region_types = self.get_attribute("region_type")
        if region_types is not None:
            region_types = set(region_types.tolist())
            assert len(region_types) == 1, f"Repertoire {self.identifier}: expected one region_type, found: {region_types}"
            return RegionType(region_types.pop())
        else:
            logging.warning(f'Repertoire {self.identifier}: region_types are not set for sequences.')
            return None

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['bnp_data']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.bnp_data = None

    def get_element_count(self):
        if self.element_count is None:
            self.load_bnp_data()
        return self.element_count

    def _make_sequence_object(self, row: dict):

        seq = ReceptorSequence(sequence_aa=row.get('sequence_aa', None),
                               sequence=row.get('sequence', None),
                               sequence_id=row.get('sequence_id', None),
                               metadata=SequenceMetadata(v_call=row.get('v_call', None),
                                                         j_call=row.get('j_call', None),
                                                         chain=row.get('chain', None),
                                                         duplicate_count=row.get('duplicate_count', None),
                                                         region_type=row.get('region_type', None),
                                                         frame_type=row.get('frame_type', None),
                                                         cell_id=row.get('cell_id', None),
                                                         custom_params={key: row[key] for key in row
                                                                        if key not in Repertoire.FIELDS}))

        return seq

    def _prepare_cell_lists(self):
        data = self.load_bnp_data()

        assert hasattr(data, 'cell_id') and getattr(data, 'cell_id') is not None, \
            f"Repertoire: cannot return receptor objects in repertoire {self.identifier} since cell_ids are not specified. " \
            f"Existing fields are: {str(dir(data))[1:-1]}"

        same_cell_lists = NumpyHelper.group_structured_array_by(data, "cell_id")
        return same_cell_lists

    def _make_receptors(self, cell_content):
        sequences = []
        for item in cell_content:
            sequences.append(self._make_sequence_object(item))
        return ReceptorBuilder.build_objects(sequences)

    def get_sequence_objects(self) -> List[ReceptorSequence]:
        """
        Lazily loads sequences from disk to reduce RAM consumption

        Returns:
            a list of ReceptorSequence objects
        """
        seqs = []

        data = self.load_bnp_data()

        for i in range(len(data)):
            seq = self._make_sequence_object(data.get_row_by_index(i))
            seqs.append(seq)

        del data
        self.bnp_data = None

        return seqs

    def load_bnp_data(self):
        if self.bnp_data is None or (isinstance(self.bnp_data, weakref.ref) and self.bnp_data() is None):
            data = bnp_read_from_file(self.data_filename, self._buffer_type)
            self.bnp_data = weakref.ref(data) if EnvironmentSettings.low_memory else data
        data = self.bnp_data() if EnvironmentSettings.low_memory else self.bnp_data
        self.element_count = len(data)
        return data

    @property
    def sequences(self):
        return self.get_sequence_objects()

    @property
    def receptors(self) -> List[Receptor]:
        """
        A property that creates a list of Receptor objects based on the cell_ids field in the following manner:
            - all sequences that have the same cell_id are grouped together
            - they are divided into groups based on the chain
            - all valid combinations of chains are created and used to make a receptor object - this means that if a cell has
              two beta (b1 and b2) and one alpha chain (a1), two receptor objects will be created: receptor1 (b1, a1), receptor2 (b2, a1)

        To avoid have multiple receptors in the same cell, use some of the preprocessing classes which could merge/eliminate multiple
        sequences. See the documentation of the preprocessing module for more information.

        Returns:
            List[Receptor]: a list of objects of Receptor class
        """
        data = self.load_bnp_data()
        receptors = []
        chains = data.chain.tolist()
        for i in range(0, len(data), 2):
            rows = data.get_rows_by_indices(i, i + 1)
            cls = ChainPair.get_chain_pair([Chain.get_chain(el) for el in chains[i:i+2]]).get_appropriate_receptor_class()
            receptors.append(cls.create_from_record(**rows))

        return receptors

    def _create_buffer_type_from_field_dict(self,
                                            type_dict: Dict[str, Any]) -> bnp.io.delimited_buffers.DelimitedBuffer:
        dataclass = make_dynamic_seq_set_dataclass(type_dict)
        return bnp.io.delimited_buffers.get_bufferclass_for_datatype(dataclass, delimiter='\t', has_header=True)



from immuneML.data_model.receptor.Receptor import Receptor
from immuneML.data_model.receptor.receptor_sequence.ReceptorSequence import ReceptorSequence


class BCReceptor(Receptor):
    FIELDS = {'heavy': object, 'light': object, 'identifier': str, 'metadata': dict}

    @classmethod
    def get_record_names(cls):
        return ['heavy_' + name for name in ReceptorSequence.get_record_names()] \
               + ['light_' + name for name in ReceptorSequence.get_record_names()] \
               + [name for name in cls.FIELDS if name not in ['heavy', 'light']]

    def __init__(self, heavy: ReceptorSequence = None, light: ReceptorSequence = None, metadata: dict = None,
                 identifier: str = None):
        super().__init__(metadata=metadata, identifier=identifier)
        self.heavy = heavy
        self.light = light

    def get_chains(self):
        return ["heavy", "light"]



import copy
import logging
from itertools import chain
from typing import Tuple

from immuneML import Constants
from immuneML.data_model.receptor.ChainPair import ChainPair
from immuneML.dsl.DefaultParamsLoader import DefaultParamsLoader
from immuneML.dsl.definition_parsers.SignalParser import check_clonal_frequency
from immuneML.dsl.symbol_table.SymbolTable import SymbolTable
from immuneML.dsl.symbol_table.SymbolType import SymbolType
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.environment.SequenceType import SequenceType
from immuneML.ml_methods.generative_models.GenerativeModel import GenerativeModel
from immuneML.simulation.SimConfig import SimConfig
from immuneML.simulation.SimConfigItem import SimConfigItem
from immuneML.simulation.implants.Signal import SignalPair
from immuneML.simulation.simulation_strategy.SimulationStrategy import SimulationStrategy
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.ReflectionHandler import ReflectionHandler


class SimulationParser:
    keyword = "simulations"

    @staticmethod
    def parse(simulations: dict, symbol_table: SymbolTable):
        for key, simulation in simulations.items():
            item, sim_dict = _parse_ligo_simulation(simulation, key, symbol_table)

            symbol_table.add(key, SymbolType.SIMULATION, item)
            simulations[key] = sim_dict

        return symbol_table, simulations


def _parse_ligo_simulation(simulation: dict, key: str, symbol_table: SymbolTable) -> Tuple[SimConfig, dict]:
    location = SimulationParser.__name__
    valid_keys = {'is_repertoire': bool, 'paired': list, 'sequence_type': str, 'p_gen_bin_count': int,
                  'simulation_strategy': str, 'sim_items': dict, 'keep_p_gen_dist': bool,
                  'remove_seqs_with_signals': bool, 'species': str, 'implanting_scaling_factor': int}

    simulation = {**DefaultParamsLoader.load("simulation", "ligo_sim_config"), **simulation}

    ParameterValidator.assert_keys(list(simulation.keys()), list(valid_keys.keys()), location, key, exclusive=True)
    for k, val_type in valid_keys.items():
        if k != 'paired':
            ParameterValidator.assert_type_and_value(simulation[k], val_type, location, k)

    sim_strategies = ReflectionHandler.all_nonabstract_subclass_basic_names(SimulationStrategy, drop_part='Strategy',
                                                                            subdirectory='simulation/simulation_strategy')

    ParameterValidator.assert_sequence_type(simulation, location)
    ParameterValidator.assert_in_valid_list(simulation['simulation_strategy'], sim_strategies, location,
                                            'simulation_strategy')

    check_sim_strategy_setup(simulation)
    sim_strategy_cls = ReflectionHandler.get_class_by_name(simulation['simulation_strategy'] + "Strategy",
                                                           "simulation/simulation_strategy")

    sim_items = []
    for sim_key, item in simulation['sim_items'].items():
        sim_item, sim_item_dict = _parse_sim_config_item(item, sim_key, symbol_table, simulation['is_repertoire'])
        strategy_supports_signals(sim_item, sim_strategy_cls)
        sim_items.append(sim_item)
        simulation['sim_items'][sim_key] = sim_item_dict

        if simulation['simulation_strategy'] == "Implanting":
            assert all(signal.is_present_custom_func is None for signal in sim_item.signals), \
                f"Simulation {key}: if using signals with custom functions, switch to RejectionSampling simulation " \
                f"strategy; Implanting does not support custom functions."

    sim_obj = SimConfig(**{**{k: v for k, v in simulation.items() if k != 'type'},
                           **{'sequence_type': SequenceType[simulation['sequence_type'].upper()],
                              "sim_items": sim_items, "identifier": key, 'simulation_strategy': sim_strategy_cls()}})

    _signal_content_matches_seq_type(sim_obj)
    _check_paired_specs(sim_obj)
    return sim_obj, simulation


def check_sim_strategy_setup(simulation: dict):
    if 'implanting_scaling_factor' in simulation and simulation['implanting_scaling_factor'] > 1:
        assert (simulation['simulation_strategy'] == 'Implanting' and simulation['keep_p_gen_dist']
                and simulation['p_gen_bin_count'] > 1), \
            (f"{SimulationParser.__name__}: 'implanting_scaling_factor' is only supported with 'Implanting' simulation "
             f"strategy when 'keep_p_gen_dist' is set to True (i.e., when importance sampling is used) and "
             f"'p_gen_bin_count' is larger than 1.")


def _check_paired_specs(sim_obj: SimConfig):
    assert sim_obj.paired is None or sim_obj.paired is False or isinstance(sim_obj.paired, list)

    if isinstance(sim_obj.paired, list):
        for paired_item in sim_obj.paired:
            assert len(paired_item) == 2
            sim_item1 = [item for item in sim_obj.sim_items if item.name == paired_item[0]]
            assert len(sim_item1) == 1, f'Sim item {paired_item[0]} is not specified under sim_items.'
            sim_item1 = sim_item1[0]

            sim_item2 = [item for item in sim_obj.sim_items if item.name == paired_item[1]]
            assert len(sim_item2) == 1, f'Sim item {paired_item[1]} is not specified under sim_items.'
            sim_item2 = sim_item2[0]

            assert sim_item1.number_of_examples == sim_item2.number_of_examples
            assert sim_item1.receptors_in_repertoire_count == sim_item2.receptors_in_repertoire_count

            ChainPair.get_chain_pair([sim_item1.generative_model.chain, sim_item2.generative_model.chain])

        assert len(list(set(chain.from_iterable(sim_obj.paired)))) == 2 * len(sim_obj.paired)
        assert all(item.name in list(chain.from_iterable(sim_obj.paired)) for item in sim_obj.sim_items)


def strategy_supports_signals(sim_item, sim_strategy_cls):
    if "Implanting" in sim_strategy_cls.__name__:
        assert all(not isinstance(sig, SignalPair) for sig in sim_item.signals), \
            "Implanting does not support having more than 1 signal per sequence. Please adjust the simulation specs."


def _parse_sim_config_item(simulation_item: dict, key: str, symbol_table: SymbolTable, is_repertoire: bool) -> Tuple[
    SimConfigItem, dict]:
    location = SimulationParser.__name__
    valid_simulation_item_keys = ["number_of_examples", "signals", "is_noise", "seed", "default_clonal_frequency",
                                  "false_positive_prob_in_receptors", "false_negative_prob_in_receptors",
                                  "sequence_len_limits",
                                  "receptors_in_repertoire_count", "generative_model", "immune_events"]

    simulation_item = {**DefaultParamsLoader.load('simulation', 'ligo_sim_config_item'), **simulation_item}

    ParameterValidator.assert_keys(simulation_item.keys(), valid_simulation_item_keys, location, key, exclusive=True)

    ParameterValidator.assert_type_and_value(simulation_item['is_noise'], bool, location, 'is_noise')
    _parse_signals(simulation_item, symbol_table, location, key)

    _validate_sequence_len_limits(simulation_item)

    for k in ['number_of_examples']:
        ParameterValidator.assert_type_and_value(simulation_item[k], int, location, k, min_inclusive=1)

    for k, val_type in zip(['receptors_in_repertoire_count', 'immune_events'], [int, dict]):
        if simulation_item[k]:
            ParameterValidator.assert_type_and_value(simulation_item[k], val_type, location, k)

    ParameterValidator.assert_all_type_and_value(simulation_item.keys(), str, location, 'immune_events')
    for k, val in simulation_item['immune_events'].items():
        assert isinstance(val, int) or isinstance(val, bool) or isinstance(val, str), \
            f"The values for immune events under {k} has to be int, bool or string, got {val} ({type(val)}."

    gen_model = _parse_generative_model(simulation_item, location)

    params = copy.deepcopy(simulation_item)
    params["signal_proportions"] = _make_signal_proportions(symbol_table, simulation_item["signals"], key)
    params["name"] = key
    params['generative_model'] = gen_model

    check_clonal_frequency(simulation_item, 'default_clonal_frequency', SimulationParser.__name__)
    if not is_repertoire:
        assert simulation_item[
                   'default_clonal_frequency'] is None, "Clonal frequency can be set only for repertoire simulation."
    if not is_repertoire and any(signal.clonal_frequency is not None for signal in params['signal_proportions'].keys()):
        logging.warning(
            f"Clonal frequency is set for some of the signals in {key}, but the simulation is not on repertoire level, "
            f"so clonal frequency parameters will not be used.")

    return SimConfigItem(**{key: val for key, val in params.items() if key not in ['signals', 'type']}), simulation_item


def _parse_signals(sim_item: dict, symbol_table: SymbolTable, location: str, key: str) -> dict:
    assert isinstance(sim_item['signals'], dict) or sim_item['signals'] is None, \
        f"Signals under {key} have to be either null or a dictionary, got: {sim_item['signals']}."

    if sim_item['signals'] is not None and bool(sim_item['signals']):

        signals = _extract_signals_from_potential_pairs(sim_item["signals"].keys())
        ParameterValidator.assert_keys(signals, symbol_table.get_keys_by_type(SymbolType.SIGNAL), location, key, False)
        for signal_key, signal_val in sim_item['signals'].items():
            ParameterValidator.assert_type_and_value(signal_val, (int, float), location, signal_key, 0., 1.)
        assert 0 <= sum(sim_item['signals'].values()) <= 1, sim_item['signals']

    else:

        sim_item['signals'] = {}

    return sim_item


def _extract_signals_from_potential_pairs(signal_keys: list):
    return list(set(chain.from_iterable([sig_key.split(Constants.SIGNAL_DELIMITER) for sig_key in signal_keys])))


def _make_signal_proportions(symbol_table: SymbolTable, signals: dict, location: str) -> dict:
    signal_proportions = {}

    for signal, proportion in signals.items():
        tmp_signals = signal.split(Constants.SIGNAL_DELIMITER)
        if len(tmp_signals) == 2:
            signal_pair = SignalPair(symbol_table.get(tmp_signals[0]), symbol_table.get(tmp_signals[1]))
            if signal_pair in signal_proportions:
                raise KeyError(f"The combinations of signals {signal_pair} that co-occur was defined multiple times. "
                               f"Please check the signals under {location}.")
            signal_proportions[signal_pair] = proportion
        elif len(tmp_signals) == 1:
            signal_proportions[symbol_table.get(signal)] = proportion
        else:
            raise RuntimeError(
                f"Couldn't parse signals listed under {location}. The keys have to be either single signal names or two "
                f"signal names separated by {Constants.SIGNAL_DELIMITER}. Got {signal} instead.")

    return signal_proportions


def _parse_generative_model(simulation_item: dict, location: str):
    ParameterValidator.assert_type_and_value(simulation_item['generative_model'], dict, location, 'generative_model')
    ParameterValidator.assert_keys_present(simulation_item['generative_model'].keys(), ['type'], location,
                                           'generative_model')

    # gen_model_classes = ReflectionHandler.all_nonabstract_subclass_basic_names(GenerativeModel, "",
    #                                                                            "ml_methods/generative_models/")

    gen_model_classes = ['OLGA', 'ExperimentalImport']  # TODO: allow other gen models to be used with Ligo as well

    ParameterValidator.assert_in_valid_list(simulation_item['generative_model']['type'], gen_model_classes, location,
                                            "generative_model:type")

    gen_model_cls = ReflectionHandler.get_class_by_name(simulation_item['generative_model']['type'],
                                                        "ml_methods/generative_models/")
    params = {key: val for key, val in simulation_item['generative_model'].items() if key != 'type'}
    gen_model = gen_model_cls.build_object(**params)

    return gen_model


def _signal_content_matches_seq_type(simulation: SimConfig):
    for sim_item in simulation.sim_items:
        for signal in sim_item.signals:
            if isinstance(signal,
                          SignalPair) and signal.signal1.is_present_custom_func is None and signal.signal2.is_present_custom_func is None:
                _motif_content_matches_seq_type(signal.signal1.motifs + signal.signal2.motifs, simulation.sequence_type)
            elif signal.is_present_custom_func is None:
                _motif_content_matches_seq_type(signal.motifs, simulation.sequence_type)


def _motif_content_matches_seq_type(motifs: list, seq_type):
    for motif_group in motifs:
        motifs = motif_group if isinstance(motif_group, list) else [motif_group]
        for motif in motifs:
            ParameterValidator.assert_all_in_valid_list(motif.get_alphabet(),
                                                        EnvironmentSettings.get_sequence_alphabet(seq_type),
                                                        SimulationParser.__name__, motif.get_alphabet())


def _validate_sequence_len_limits(sim_item: dict):
    ParameterValidator.assert_keys(sim_item['sequence_len_limits'].keys(), ['min', 'max'], SimulationParser.__name__,
                                   'sequence_len_limits')
    for key in ['min', 'max']:
        ParameterValidator.assert_type_and_value(sim_item['sequence_len_limits'][key], int, SimulationParser.__name__,
                                                 f'sequence_len_limits:{key}', -1)

    assert sim_item['sequence_len_limits']['min'] <= sim_item['sequence_len_limits']['max'] or \
           sim_item['sequence_len_limits']['max'] == -1, \
        (f"Under sequence_len_limits, min has to be less or equal to max value, if max is not -1 (max=-1 -> max is not "
         f"used): {sim_item['sequence_len_limits']}")



import shutil
from pathlib import Path

import numpy as np
from olga import load_model
from sonia.sequence_generation import SequenceGeneration
from sonnia.sonnia import SoNNia as InternalSoNNia

from immuneML.data_model.bnp_util import write_yaml, read_yaml
from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.data_model.dataset.SequenceDataset import SequenceDataset
from immuneML.data_model.receptor.RegionType import RegionType
from immuneML.data_model.receptor.receptor_sequence.Chain import Chain
from immuneML.data_model.receptor.receptor_sequence.ReceptorSequence import ReceptorSequence
from immuneML.data_model.receptor.receptor_sequence.SequenceMetadata import SequenceMetadata
from immuneML.environment.SequenceType import SequenceType
from immuneML.ml_methods.generative_models.GenerativeModel import GenerativeModel
from immuneML.ml_methods.generative_models.OLGA import OLGA
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder


class SoNNia(GenerativeModel):
    """
    SoNNia models the selection process of T and B cell receptor repertoires. It is based on the SoNNia Python package.
    It supports SequenceDataset as input, but not RepertoireDataset.

    Original publication:
    Isacchini, G., Walczak, A. M., Mora, T., & Nourmohammad, A. (2021). Deep generative selection models of T and B
    cell receptor repertoires with soNNia. Proceedings of the National Academy of Sciences, 118(14), e2023141118.
    https://doi.org/10.1073/pnas.2023141118

    Specification arguments:

    - chain (str)

    - batch_size (int)

    - epochs (int)

    - deep (bool)

    - include_joint_genes (bool)

    - n_gen_seqs (int)

    - custom_model_path (str)

    - default_model_name (str)

    """

    @classmethod
    def load_model(cls, path: Path):
        assert path.exists(), f"{cls.__name__}: {path} does not exist."

        model_overview_file = path / 'model_overview.yaml'

        for file in [model_overview_file]:
            assert file.exists(), f"{cls.__name__}: {file} is not a file."

        model_overview = read_yaml(model_overview_file)
        sonnia = SoNNia(**{k: v for k, v in model_overview.items() if k != 'type'})
        sonnia._model = InternalSoNNia(load_dir=path)
        return sonnia

    def __init__(self, chain=None, batch_size: int = None, epochs: int = None, deep: bool = False, name: str = None,
                 default_model_name: str = None, n_gen_seqs: int = None, include_joint_genes: bool = True,
                 custom_model_path: str = None):
        if chain is not None:
            super().__init__(chain)
        elif default_model_name is not None:
            super().__init__(chain=Chain.get_chain(default_model_name[-3:]))
        self.epochs = epochs
        self.batch_size = int(batch_size)
        self.deep = deep
        self.include_joint_genes = include_joint_genes
        self.n_gen_seqs = n_gen_seqs
        self._model = None
        self.name = name
        self.default_model_name = default_model_name
        if custom_model_path is None or custom_model_path == '':
            self._model_path = Path(
                load_model.__file__).parent / f"default_models/{OLGA.DEFAULT_MODEL_FOLDER_MAP[self.default_model_name]}"
        else:
            self._model_path = custom_model_path

    def fit(self, dataset: Dataset, path: Path = None):
        print_log(f"{SoNNia.__name__}: fitting a selection model...", True)

        data = dataset.get_attributes(['sequence_aa', 'v_call', 'j_call'], as_list=True)
        data_seqs = [[data['sequence_aa'][i], data['v_call'][i], data['j_call'][i]]
                     for i in range(len(data['sequence_aa']))]

        self._model = InternalSoNNia(data_seqs=data_seqs,
                                     gen_seqs=[],
                                     custom_pgen_model=self._model_path,
                                     vj=self.chain in [Chain.ALPHA, Chain.KAPPA, Chain.LIGHT],
                                     include_joint_genes=self.include_joint_genes,
                                     include_indep_genes=not self.include_joint_genes)

        self._model.add_generated_seqs(num_gen_seqs=self.n_gen_seqs, custom_model_folder=self._model_path)

        self._model.infer_selection(epochs=self.epochs, batch_size=self.batch_size, verbose=1)

        print_log(f"{SoNNia.__name__}: selection model fitted.", True)

    def is_same(self, model) -> bool:
        raise NotImplementedError

    def generate_sequences(self, count: int, seed: int, path: Path, sequence_type: SequenceType, compute_p_gen: bool):
        gen_model = SequenceGeneration(self._model)
        sequences = gen_model.generate_sequences_post(count)
        return SequenceDataset.build_from_objects(sequences=[ReceptorSequence(sequence_aa=seq[0], sequence=seq[3],
                                                                              metadata=SequenceMetadata(
                                                                                  v_call=seq[1],
                                                                                  j_call=seq[2],
                                                                                  region_type=RegionType.IMGT_JUNCTION.name))
                                                             for seq in sequences],
                                                  file_size=len(sequences), path=PathBuilder.build(path),
                                                  name='SoNNiaDataset')

    def compute_p_gens(self, sequences, sequence_type: SequenceType) -> np.ndarray:
        raise NotImplementedError

    def compute_p_gen(self, sequence: dict, sequence_type: SequenceType) -> float:
        raise NotImplementedError

    def can_compute_p_gens(self) -> bool:
        return False

    def can_generate_from_skewed_gene_models(self) -> bool:
        return False

    def generate_from_skewed_gene_models(self, v_genes: list, j_genes: list, seed: int, path: Path,
                                         sequence_type: SequenceType, batch_size: int, compute_p_gen: bool):
        raise NotImplementedError

    def save_model(self, path: Path) -> Path:
        PathBuilder.build(path / 'model')

        write_yaml(path / 'model/model_overview.yaml', {'type': 'SoNNia', 'chain': self.chain.name,
                                                        **{k: v for k, v in vars(self).items()
                                                           if k not in ['_model', 'chain', '_model_path']}})
        self._model.save_model(path / 'model')

        return Path(shutil.make_archive(str(path / 'trained_model'), "zip", str(path / 'model'))).absolute()



import hashlib
from inspect import signature
from pathlib import Path
from typing import Tuple

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.dsl.DefaultParamsLoader import DefaultParamsLoader
from immuneML.dsl.definition_parsers.PreprocessingParser import PreprocessingParser
from immuneML.dsl.instruction_parsers.LabelHelper import LabelHelper
from immuneML.dsl.symbol_table.SymbolTable import SymbolTable
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.example_weighting.ExampleWeightingStrategy import ExampleWeightingStrategy
from immuneML.hyperparameter_optimization.HPSetting import HPSetting
from immuneML.hyperparameter_optimization.config.LeaveOneOutConfig import LeaveOneOutConfig
from immuneML.hyperparameter_optimization.config.ManualSplitConfig import ManualSplitConfig
from immuneML.hyperparameter_optimization.config.ReportConfig import ReportConfig
from immuneML.hyperparameter_optimization.config.SplitConfig import SplitConfig
from immuneML.hyperparameter_optimization.config.SplitType import SplitType
from immuneML.ml_metrics.ClassificationMetric import ClassificationMetric
from immuneML.reports.train_ml_model_reports.TrainMLModelReport import TrainMLModelReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.ReflectionHandler import ReflectionHandler
from immuneML.workflows.instructions.TrainMLModelInstruction import TrainMLModelInstruction


class TrainMLModelParser:

    def parse(self, key: str, instruction: dict, symbol_table: SymbolTable, path: Path = None) -> TrainMLModelInstruction:

        valid_keys = ["assessment", "selection", "dataset", "strategy", "labels", "metrics", "settings", "number_of_processes", "type", "reports",
                      "optimization_metric", "refit_optimal_model", "example_weighting", "export_all_ml_settings"]
        ParameterValidator.assert_type_and_value(instruction['settings'], list, TrainMLModelParser.__name__, 'settings')
        ParameterValidator.assert_keys(list(instruction.keys()), valid_keys, TrainMLModelParser.__name__, "TrainMLModel")
        ParameterValidator.assert_type_and_value(instruction['refit_optimal_model'], bool, TrainMLModelParser.__name__, 'refit_optimal_model')
        ParameterValidator.assert_type_and_value(instruction['export_all_ml_settings'], bool, TrainMLModelParser.__name__, 'export_all_ml_settings')
        ParameterValidator.assert_type_and_value(instruction['metrics'], list, TrainMLModelParser.__name__, 'metrics')
        ParameterValidator.assert_type_and_value(instruction['optimization_metric'], str, TrainMLModelParser.__name__, 'optimization_metric')
        ParameterValidator.assert_type_and_value(instruction['number_of_processes'], int, TrainMLModelParser.__name__, 'number_of_processes')
        ParameterValidator.assert_type_and_value(instruction['strategy'], str, TrainMLModelParser.__name__, 'strategy')
        if instruction["reports"] is not None:
            ParameterValidator.assert_type_and_value(instruction['reports'], list, TrainMLModelParser.__name__, 'reports')

        if instruction["example_weighting"] is not None:
            ParameterValidator.assert_type_and_value(instruction['example_weighting'], str, TrainMLModelParser.__name__, 'example_weighting')

        settings = self._parse_settings(instruction, symbol_table)
        dataset = symbol_table.get(instruction["dataset"])
        label_config = LabelHelper.create_label_config(instruction["labels"], dataset, TrainMLModelParser.__name__, key)
        assessment = self._parse_split_config(key, instruction, "assessment", symbol_table, len(settings), label_config)
        selection = self._parse_split_config(key, instruction, "selection", symbol_table, len(settings), label_config)
        assessment, selection = self._update_split_configs(assessment, selection, dataset)
        strategy = ReflectionHandler.get_class_by_name(instruction["strategy"], "hyperparameter_optimization/")
        metrics = {ClassificationMetric.get_metric(metric) for metric in instruction["metrics"]}
        optimization_metric = ClassificationMetric.get_metric(instruction["optimization_metric"].upper())
        metric_search_criterion = ClassificationMetric.get_search_criterion(optimization_metric)
        path = self._prepare_path(instruction)
        context = self._prepare_context(instruction, symbol_table)
        reports = self._prepare_reports(instruction["reports"], symbol_table)
        example_weighting = self._prepare_example_weighting(instruction, symbol_table)

        hp_instruction = TrainMLModelInstruction(dataset=dataset, hp_strategy=strategy(settings, metric_search_criterion),
                                                 hp_settings=settings, assessment=assessment, selection=selection, metrics=metrics,
                                                 optimization_metric=optimization_metric, refit_optimal_model=instruction['refit_optimal_model'],
                                                 label_configuration=label_config, path=path, context=context,
                                                 number_of_processes=instruction["number_of_processes"], reports=reports,
                                                 example_weighting=example_weighting, export_all_ml_settings=instruction['export_all_ml_settings'],
                                                 name=key)

        return hp_instruction

    def _update_split_configs(self, assessment: SplitConfig, selection: SplitConfig, dataset: Dataset) -> Tuple[SplitConfig, SplitConfig]:

        if assessment.split_strategy == SplitType.LOOCV:
            assessment.split_count = dataset.get_example_count()
            train_val_example_count = assessment.split_count - 1
        elif assessment.split_strategy == SplitType.K_FOLD or assessment.split_strategy.STRATIFIED_K_FOLD:
            train_val_example_count = int(dataset.get_example_count() * (assessment.split_count - 1) / assessment.split_count)
        else:
            train_val_example_count = int(dataset.get_example_count() * assessment.training_percentage)

        if selection.split_strategy == SplitType.LOOCV:
            selection.split_count = train_val_example_count

        return assessment, selection

    def _prepare_reports(self, reports: list, symbol_table: SymbolTable) -> dict:
        if reports is not None:
            ParameterValidator.assert_type_and_value(reports, list, TrainMLModelParser.__name__, "reports")
            report_objects = {report_id: symbol_table.get(report_id) for report_id in reports}
            ParameterValidator.assert_all_type_and_value(report_objects.values(), TrainMLModelReport, TrainMLModelParser.__name__, 'reports')
            return report_objects
        else:
            return {}

    def _prepare_context(self, instruction: dict, symbol_table: SymbolTable):
        return {"dataset": symbol_table.get(instruction["dataset"])}

    def _parse_settings(self, instruction: dict, symbol_table: SymbolTable) -> list:
        try:
            settings = []
            for index, setting in enumerate(instruction["settings"]):
                if "preprocessing" in setting and setting["preprocessing"] is not None:
                    ParameterValidator.assert_type_and_value(setting["preprocessing"], str, TrainMLModelParser.__name__, f'settings: {index+1}. '
                                                                                                                         f'element: preprocessing')
                    if symbol_table.contains(setting["preprocessing"]):
                        preprocessing_sequence = symbol_table.get(setting["preprocessing"])
                        preproc_name = setting["preprocessing"]
                        if not all(preproc.keeps_example_count() for preproc in preprocessing_sequence):
                            raise ValueError(f"{TrainMLModelParser.__name__}: preprocessing sequence {preproc_name} includes preprocessing that "
                                             f"change the number of examples at runtime and as such cannot be used with this instruction. See the "
                                             f"documentation for the preprocessing or alternatively use them with other instructions.")
                    else:
                        raise KeyError(f"{TrainMLModelParser.__name__}: preprocessing was set in the TrainMLModel instruction to value "
                                       f"{setting['preprocessing']}, but no such preprocessing was defined in the specification under "
                                       f"definitions: {PreprocessingParser.keyword}.")
                else:
                    setting["preprocessing"] = None
                    preprocessing_sequence = []
                    preproc_name = None

                ParameterValidator.assert_keys(setting.keys(), ["preprocessing", "ml_method", "encoding"], TrainMLModelParser.__name__,
                                               f"settings, {index + 1}. entry")

                encoder = symbol_table.get(setting["encoding"]).build_object(symbol_table.get(instruction["dataset"]),
                                                                             **symbol_table.get_config(setting["encoding"])["encoder_params"])\
                    .set_context({"dataset": symbol_table.get(instruction['dataset'])})

                ml_method = symbol_table.get(setting["ml_method"])
                ml_method.check_encoder_compatibility(encoder)

                s = HPSetting(encoder=encoder,
                              encoder_name=setting["encoding"],
                              encoder_params=symbol_table.get_config(setting["encoding"])["encoder_params"],
                              ml_method=ml_method,
                              ml_method_name=setting["ml_method"],
                              ml_params=symbol_table.get_config(setting["ml_method"]),
                              preproc_sequence=preprocessing_sequence, preproc_sequence_name=preproc_name)
                settings.append(s)
            return settings
        except KeyError as key_error:
            raise KeyError(f"{TrainMLModelParser.__name__}: parameter {key_error.args[0]} was not defined under settings in TrainMLModel instruction.")

    def _prepare_path(self, instruction: dict) -> Path:
        if "path" in instruction:
            path = Path(instruction["path"]).absolute()
        else:
            path = EnvironmentSettings.default_analysis_path / hashlib.md5(str(instruction).encode()).hexdigest()

        return path

    def _parse_split_config(self, instruction_key, instruction: dict, split_key: str, symbol_table: SymbolTable, settings_count: int,
                            label_config: LabelConfiguration) -> SplitConfig:

        try:

            default_params = DefaultParamsLoader.load("instructions/", SplitConfig.__name__)
            report_config_input = self._prepare_report_config(instruction_key, instruction, split_key, symbol_table)
            instruction[split_key] = {**default_params, **instruction[split_key]}

            split_strategy = SplitType[instruction[split_key]["split_strategy"].upper()]
            training_percentage = float(instruction[split_key]["training_percentage"]) if split_strategy == SplitType.RANDOM else -1

            if split_strategy == SplitType.RANDOM and training_percentage == 1 and settings_count > 1:
                raise ValueError(f"{TrainMLModelParser.__name__}: all data under {instruction_key}/{split_key} was specified to be used for "
                                 f"training, but {settings_count} settings were specified for evaluation. Please define a test/validation set by "
                                 f"reducing the training percentage (e.g., to 0.7) or use only one hyperparameter setting to run the analysis.")

            if split_strategy == SplitType.STRATIFIED_K_FOLD and len(label_config.get_labels_by_name()) != 1:
                raise ValueError(f"{TrainMLModelParser.__name__}: Stratified k-fold cross-validation cannot be used when "
                                 f"{len(label_config.get_labels_by_name())} labels are specified. It support only one label (and multiple classes).")

            if split_strategy == SplitType.MANUAL:
                ParameterValidator.assert_keys(keys=instruction[split_key]["manual_config"].keys(),
                                               valid_keys=["train_metadata_path", "test_metadata_path"],
                                               location=TrainMLModelParser.__name__, parameter_name="manual_config", exclusive=True)

                ParameterValidator.assert_valid_tabular_file(instruction[split_key]["manual_config"]["train_metadata_path"],
                                                             location=TrainMLModelParser.__name__,
                                                             parameter_name="train_metadata_path")

                ParameterValidator.assert_valid_tabular_file(instruction[split_key]["manual_config"]["test_metadata_path"],
                                                             location=TrainMLModelParser.__name__,
                                                             parameter_name="test_metadata_path")

            return SplitConfig(split_strategy=split_strategy,
                               split_count=int(instruction[split_key]["split_count"]),
                               training_percentage=training_percentage,
                               reports=ReportConfig(**report_config_input),
                               manual_config=ManualSplitConfig(**instruction[split_key]["manual_config"]) if "manual_config" in instruction[split_key] else None,
                               leave_one_out_config=LeaveOneOutConfig(**instruction[split_key]["leave_one_out_config"])
                               if "leave_one_out_config" in instruction[split_key] else None)

        except KeyError as key_error:
            raise KeyError(f"{TrainMLModelParser.__name__}: parameter {key_error.args[0]} was not defined under {split_key}.")

    def _prepare_report_config(self, instruction_key, instruction, split_key, symbol_table):
        if "reports" in instruction[split_key] and len(instruction[split_key]["reports"]) > 0:
            location = f"{instruction_key}/{split_key}/reports"
            report_types = list(signature(ReportConfig).parameters.keys())
            ParameterValidator.assert_all_in_valid_list(instruction[split_key]["reports"].keys(), report_types,
                                                        location, "reports")

            for report_type in instruction[split_key]["reports"]:
                ParameterValidator.assert_type_and_value(instruction[split_key]["reports"][report_type], list, f"{location}/{report_type}",
                                                         report_type)

            report_config_input = {report_type: {report_id: symbol_table.get(report_id) for report_id in instruction[split_key]["reports"][report_type]}
                                   for report_type in instruction[split_key]["reports"]}
        else:
            report_config_input = {}

        return report_config_input


    def _prepare_example_weighting(self, instruction: dict, symbol_table: SymbolTable) -> ExampleWeightingStrategy:
        example_weighting = instruction["example_weighting"]

        if example_weighting is not None:
            weighting_strategy_cls = symbol_table.get(example_weighting)

            weighting_strategy_object = weighting_strategy_cls.build_object(symbol_table.get(instruction["dataset"]),
                                                                            **symbol_table.get_config(example_weighting)[
                                                                             "example_weighting_params"]).set_context({"dataset": symbol_table.get(instruction['dataset'])})


            ParameterValidator.assert_type_and_value(weighting_strategy_object, ExampleWeightingStrategy,  TrainMLModelParser.__name__, "example_weighting")
            return weighting_strategy_object
        else:
            return None


from immuneML.data_model.receptor.receptor_sequence.ReceptorSequence import ReceptorSequence
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.kmer_frequency.sequence_encoding.SequenceEncodingStrategy import SequenceEncodingStrategy
from immuneML.environment.Constants import Constants
from immuneML.environment.EnvironmentSettings import EnvironmentSettings


class IdentitySequenceEncoder(SequenceEncodingStrategy):

    @staticmethod
    def encode_sequence(sequence: ReceptorSequence, params: EncoderParams):
        """
        Encodes a ReceptorSequence based on information from within the ReceptorSequence and SequenceMetadata
        instances. This allows for looking at frequency for whole sequences, with flexible definition of what a unique
        whole sequence is.
        :param sequence: ReceptorSequence
        :param params: EncoderParams (params["model"]["sequence"] and params["model"]["metadata_fields_to_include"] are
                        used)
        :return: list with only single feature
        """

        res = []
        sequence_type = params.model.get('sequence_type', EnvironmentSettings.sequence_type)
        if params.model.get("sequence", True):
            res.append(sequence.get_sequence(sequence_type))

        for field in params.model.get("metadata_fields_to_include", []):
            if sequence.metadata is None:
                res.append("unknown")
            else:
                res.append(getattr(sequence.metadata, field))

        return [Constants.FEATURE_DELIMITER.join(res)]

    @staticmethod
    def get_feature_names(params: EncoderParams):
        res = []
        if params.model.get("sequence", True):
            res.append("sequence")
        for field in params.model.get("metadata_fields_to_include", []):
            res.append(field)
        return res



from pathlib import Path

import numpy as np
import pandas as pd

from immuneML.caching.CacheHandler import CacheHandler
from immuneML.data_model.dataset.RepertoireDataset import RepertoireDataset
from immuneML.pairwise_repertoire_comparison.ComparisonData import ComparisonData
from immuneML.util.Logger import log
from immuneML.util.PathBuilder import PathBuilder


class PairwiseRepertoireComparison:

    @log
    def __init__(self, matching_columns: list, item_columns: list, path: Path, sequence_batch_size: int):
        self.matching_columns = matching_columns
        self.item_columns = item_columns
        self.path = PathBuilder.build(path)
        self.sequence_batch_size = sequence_batch_size
        self.comparison_data = None
        self.comparison_fn = None

    @log
    def create_comparison_data(self, dataset: RepertoireDataset) -> ComparisonData:

        comparison_data = ComparisonData(dataset.get_repertoire_ids(), self.matching_columns, self.sequence_batch_size, self.path)
        comparison_data.process_dataset(dataset)

        return comparison_data

    def prepare_caching_params(self, dataset: RepertoireDataset):
        return (
            ("dataset_identifier", dataset.identifier),
            ("item_attributes", self.item_columns)
        )

    def compare(self, dataset: RepertoireDataset, comparison_fn, comparison_fn_name):
        return CacheHandler.memo_by_params((("dataset_identifier", dataset.identifier),
                                            "pairwise_comparison",
                                            ("comparison_fn", comparison_fn_name)),
                                           lambda: self.compare_repertoires(dataset, comparison_fn))

    def memo_by_params(self, dataset: RepertoireDataset):
        comparison_data = CacheHandler.memo_by_params(self.prepare_caching_params(dataset), lambda: self.create_comparison_data(dataset))
        return comparison_data

    @log
    def compare_repertoires(self, dataset: RepertoireDataset, comparison_fn):
        self.comparison_data = self.memo_by_params(dataset)
        repertoire_count = dataset.get_example_count()
        comparison_result = np.zeros([repertoire_count, repertoire_count])
        repertoire_identifiers = dataset.get_repertoire_ids()

        for index1 in range(repertoire_count):
            repertoire_vector_1 = self.comparison_data.get_repertoire_vector(repertoire_identifiers[index1])
            for index2 in range(index1, repertoire_count):
                repertoire_vector_2 = self.comparison_data.get_repertoire_vector(repertoire_identifiers[index2])
                comparison_result[index1, index2] = comparison_fn(repertoire_vector_1, repertoire_vector_2)
                comparison_result[index2, index1] = comparison_result[index1, index2]

        comparison_df = pd.DataFrame(comparison_result, columns=repertoire_identifiers, index=repertoire_identifiers)

        return comparison_df

    def prepare_paralellization_arguments(self, repertoire_count: int, repertoire_identifiers: list, comparison_result):

        arguments = []

        for index1 in range(repertoire_count):
            comparison_result[index1, index1] = 1
            rep1 = repertoire_identifiers[index1]
            for index2 in range(index1+1, repertoire_count):
                rep2 = repertoire_identifiers[index2]
                arguments.append((rep1, rep2))

        return arguments



import logging


class Label:

    def __init__(self, name: str, values: list, auxiliary_label_names: list = None, positive_class=None):
        self.name = name
        self._values = values
        self.auxiliary_label_names = auxiliary_label_names
        self._positive_class = positive_class

    def __str__(self):
        return f"label {self.name} ({', '.join([str(val) for val in self.values])})"

    @property
    def positive_class(self):
        if self._positive_class is not None:
            return self._positive_class
        else:
            assert self._values is not None, f"Label: cannot access positive class for label {self.name} " \
                                             f"when neither 'positive_class' nor 'values' are set."

            positive_class = sorted(self._values)[0]
            logging.info(f"Label: No positive class was set for label {self.name}. "
                         f"Assuming default positive class '{positive_class}'.")

            return positive_class

    @property
    def values(self):
        '''
        Make sure the positive class is listed last
        This is needed for compatibility with `~immuneML.ml_methods.util.Util.binarize_label_classes`
        '''
        if self._positive_class is not None:
            negative_classes = sorted(self._values)
            negative_classes.remove(self._positive_class)
            return negative_classes + [self._positive_class]
        else:
            return sorted(self._values)

    def get_binary_negative_class(self):
        assert len(self._values) == 2, f"Label: binary negative class was requested for label {self.name} but this label contains {len(self._values)} classes: {self.values}"
        return [val for val in self._values if val != self.positive_class][0]

    def get_desc_for_storage(self):
        return {key.lstrip("_"): value for key, value in vars(self).items()}



import abc

import numpy as np
import pandas as pd
from scipy import sparse

from immuneML.caching.CacheHandler import CacheHandler
from immuneML.data_model.encoded_data.EncodedData import EncodedData
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.util.EncoderHelper import EncoderHelper
from immuneML.util.ReflectionHandler import ReflectionHandler


class EvennessProfileEncoder(DatasetEncoder):
    """
    The EvennessProfileEncoder class encodes a repertoire based on the clonal frequency distribution. The evenness for
    a given repertoire is defined as follows:

    .. math::

        ^{\\alpha} \\mathrm{E}(\\mathrm{f})=\\frac{\\left(\\sum_{\\mathrm{i}=1}^{\\mathrm{n}} \\mathrm{f}_{\\mathrm{i}}^{\\alpha}\\right)^{\\frac{1}{1-\\alpha}}}{\\mathrm{n}}

    That is, it is the exponential of Renyi entropy at a given alpha divided by the species richness, or number of unique
    sequences.

    Reference: Greiff et al. (2015). A bioinformatic framework for immune repertoire diversity profiling enables detection of immunological
    status. Genome Medicine, 7(1), 49. `doi.org/10.1186/s13073-015-0169-8 <https://doi.org/10.1186/s13073-015-0169-8>`_


    Specification arguments:

    - min_alpha (float): minimum alpha value to use

    - max_alpha (float): maximum alpha value to use

    - dimension (int): dimension of output evenness profile vector, or the number of alpha values to linearly space
      between min_alpha and max_alpha

    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

            my_evenness_profile:
                EvennessProfile:
                    min_alpha: 0
                    max_alpha: 10
                    dimension: 51


    """

    STEP_ENCODED = "encoded"
    STEP_VECTORIZED = "vectorized"

    dataset_mapping = {
        "RepertoireDataset": "EvennessProfileRepertoireEncoder",
    }

    def __init__(self, min_alpha: float, max_alpha: float, dimension: int, name: str = None):
        self.min_alpha = min_alpha
        self.max_alpha = max_alpha
        self.dimension = dimension
        self.name = name

    @staticmethod
    def _prepare_parameters(min_alpha: float, max_alpha: float, dimension: int, name: str = None):

        return {
            "min_alpha": min_alpha,
            "max_alpha": max_alpha,
            "dimension": dimension,
            "name": name
        }

    @staticmethod
    def build_object(dataset=None, **params):
        EncoderHelper.check_dataset_type_available_in_mapping(dataset, EvennessProfileEncoder)

        prepared_params = EvennessProfileEncoder._prepare_parameters(**params)
        encoder = ReflectionHandler.get_class_by_name(EvennessProfileEncoder.dataset_mapping[dataset.__class__.__name__],
                                                          "evenness_profile/")(**prepared_params)

        return encoder

    def encode(self, dataset, params: EncoderParams):

        encoded_dataset = CacheHandler.memo_by_params(self._prepare_caching_params(dataset, params),
                                                      lambda: self._encode_new_dataset(dataset, params))

        return encoded_dataset

    def _prepare_caching_params(self, dataset, params: EncoderParams, step: str = ""):
        return (("example_identifiers", tuple(dataset.get_example_ids())),
                ("dataset_metadata", dataset.metadata_file if hasattr(dataset, "metadata_file") else None),
                ("dataset_type", dataset.__class__.__name__),
                ("labels", tuple(params.label_config.get_labels_by_name())),
                ("encoding", EvennessProfileEncoder.__name__),
                ("learn_model", params.learn_model),
                ("step", step),
                ("encoding_params", tuple(vars(self).items())))

    def _encode_data(self, dataset, params: EncoderParams) -> EncodedData:

        encoded_example_list, example_ids, encoded_labels = CacheHandler.memo_by_params(
            self._prepare_caching_params(dataset, params, EvennessProfileEncoder.STEP_ENCODED),
            lambda: self._encode_examples(dataset, params))

        vectorized_examples = CacheHandler.memo_by_params(
            self._prepare_caching_params(dataset, params, EvennessProfileEncoder.STEP_VECTORIZED),
            lambda: self._vectorize_encoded(examples=encoded_example_list))

        feature_names = list(range(self.dimension))

        feature_annotations = pd.DataFrame({"feature": feature_names})

        encoded_data = EncodedData(examples=vectorized_examples,
                                   labels=encoded_labels,
                                   feature_names=feature_names,
                                   example_ids=example_ids,
                                   feature_annotations=feature_annotations,
                                   example_weights=EncoderHelper.get_example_weights_by_identifiers(dataset, example_ids),
                                   encoding=EvennessProfileEncoder.__name__)

        return encoded_data

    @abc.abstractmethod
    def _encode_new_dataset(self, dataset, params: EncoderParams):
        pass

    @abc.abstractmethod
    def _encode_examples(self, dataset, params: EncoderParams):
        pass

    def _vectorize_encoded(self, examples: list):

        vectorized_examples = sparse.csr_matrix(np.vstack(examples))

        return vectorized_examples



import abc

from immuneML.workflows.steps.StepParams import StepParams


class Step(metaclass=abc.ABCMeta):
    """
    This class encapsulates steps in the analysis which will likely be often used, such as:
        - dataset encoding
        - training of machine learning models
        - signal implanting in repertoires without any signals etc.

    For a custom analysis which is not likely to be repeated for different settings
    (e.g. such as with a different encoding), create a custom class inheriting
    AbstractProcess from workflows.processes package.
    """

    @staticmethod
    @abc.abstractmethod
    def run(input_params: StepParams = None):
        pass



import dataclasses
from dataclasses import fields as get_fields
from enum import Enum
from itertools import chain
from pathlib import Path
from typing import Dict, Any

import bionumpy as bnp
import numpy as np
import yaml
from bionumpy import EncodedArray
from bionumpy.bnpdataclass import bnpdataclass

from immuneML.data_model.SequenceSet import SequenceSet
from immuneML.data_model.receptor.receptor_sequence.Chain import Chain
from immuneML.util.ReflectionHandler import ReflectionHandler


def bnp_write_to_file(filename: Path, bnp_object):
    buffer_type = bnp.io.delimited_buffers.get_bufferclass_for_datatype(type(bnp_object), delimiter='\t',
                                                                        has_header=True)
    with bnp.open(str(filename), 'w', buffer_type=buffer_type) as file:
        file.write(bnp_object)


def bnp_read_from_file(filename: Path, buffer_type: bnp.io.delimited_buffers.DelimitedBuffer = None, dataclass=None):
    if buffer_type is None:
        buffer_type = bnp.io.delimited_buffers.get_bufferclass_for_datatype(dataclass, delimiter='\t', has_header=True)
    with bnp.open(str(filename), buffer_type=buffer_type) as file:
        return file.read()  # TODO: fix - throws error when empty file (no lines after header)


def write_yaml(filename: Path, yaml_dict):
    with filename.open('w') as file:
        yaml.dump(yaml_dict, file)


def read_yaml(filename: Path) -> dict:
    with filename.open("r") as file:
        content = yaml.safe_load(file)
    return load_type_dict(content)


def load_type_dict(full_dict: dict) -> dict:
    if 'type_dict' in full_dict:
        full_dict['type_dict'] = {key: SequenceSet.STR_TO_TYPE[val] for key, val in full_dict['type_dict'].items()}
    return full_dict


def build_dynamic_bnp_dataclass(all_fields_dict: Dict[str, Any]):
    sequence_field_names = {**{field.name: field.type for field in dataclasses.fields(SequenceSet)},
                            **SequenceSet.additional_fields_with_types()}
    types = {}
    for key, value in all_fields_dict.items():
        if key in sequence_field_names:
            field_type = sequence_field_names[key]
        else:
            field_type = get_field_type_from_values(value)
        types[key] = field_type

    dc = make_dynamic_seq_set_dataclass(type_dict=types)
    return dc, types


def build_dynamic_bnp_dataclass_obj(all_fields_dict: Dict[str, Any]):
    dc, types = build_dynamic_bnp_dataclass(all_fields_dict)
    all_fields_dict = add_neutral_values(all_fields_dict, types)
    all_fields_dict = convert_to_expected_types(all_fields_dict, types)
    return dc(**all_fields_dict), types


def convert_to_expected_types(all_fields_dict, types) -> dict:
    for field_name, field_type in types.items():
        all_fields_dict[field_name] = [field_type(val) for val in all_fields_dict[field_name]]

    return all_fields_dict


def prepare_values_for_bnp(field_values: dict, types: dict) -> dict:
    values = add_neutral_values(field_values, types)
    values = convert_enums_to_str(values)
    return values


def add_neutral_values(field_values: dict, types: dict) -> dict:
    return {
        field: [val if val is not None else SequenceSet.get_neutral_value(types[field]) for val in values]
        for field, values in field_values.items()
    }


def convert_enums_to_str(field_values: dict) -> dict:
    return {
        field: [val.to_string() if isinstance(val, Enum) and hasattr(val, 'to_string') else val for val in values]
        for field, values in field_values.items()
    }


def make_dynamic_seq_set_dataclass(type_dict: Dict[str, Any]):
    bnp_dc = bnpdataclass(dataclasses.make_dataclass('DynamicSequenceSet', fields=type_dict.items()))

    methods = {'get_row_by_index': lambda self, index: get_row_by_index(self, index),
               'get_single_row_value': lambda self, attr_name: get_single_row_value(self, attr_name),
               'to_dict': lambda self: {field: getattr(self, field).tolist() for field in type_dict.keys()},
               'get_rows_by_indices': lambda self, index1, index2: get_rows_by_indices(self, index1, index2)
               }

    for method_name, method in methods.items():
        setattr(bnp_dc, method_name, method)

    return bnp_dc


def get_receptor_attributes_for_bnp(receptors, receptor_dc, types) -> dict:
    field_vals = {}
    for field_obj in dataclasses.fields(receptor_dc):
        if receptors[0].metadata and field_obj.name in receptors[0].metadata:
            field_vals[field_obj.name] = list(chain.from_iterable((receptor.get_attribute(field_obj.name), receptor.get_attribute(field_obj.name)) for receptor in receptors))
        elif field_obj.name == 'identifier':
            field_vals[field_obj.name] = list(chain.from_iterable((receptor.identifier, receptor.identifier) for receptor in receptors))
        else:
            field_vals[field_obj.name] = list(chain.from_iterable([receptor.get_chain(ch).get_attribute(field_obj.name)
                                                                   for ch in receptor.get_chains()]
                                                                  for receptor in receptors))
        field_vals[field_obj.name] = [el.to_string() if hasattr(el, 'to_string') else el
                                      for el in field_vals[field_obj.name]]
    return add_neutral_values(field_vals, types)


def make_dynamic_seq_set_from_objs(objs: list):
    fields = list(set(chain.from_iterable(obj.get_all_attribute_names() for obj in objs)))
    all_fields_dict = {}

    for field in fields:
        vals = [obj.get_attribute(field) for obj in objs]
        if any(isinstance(el, list) for el in vals):
            vals = list(chain.from_iterable(vals))
        all_fields_dict[field] = vals

    return build_dynamic_bnp_dataclass(all_fields_dict)


def get_field_type_from_values(values):
    t = None
    if isinstance(values, np.ndarray):
        t = type(values[0].item())
    elif len(values) == 0:
        t = str
    elif values[0] is not None:
        if issubclass(type(values[0]), Enum):
            t = str
        else:
            t = type(values[0])
    else:
        proper_values = [v for v in values if v is not None]
        if len(proper_values) > 0:
            t = type(proper_values[0])
        else:
            t = str

    return t


def get_row_by_index(self, index) -> dict:
    field_names = [f.name for f in dataclasses.fields(self)]
    d = dict()

    for field_name in field_names:
        field_value = getattr(self, field_name)[index]

        if isinstance(field_value, EncodedArray):
            field_value = field_value.to_string()
        else:
            field_value = field_value.item()

        d[field_name] = field_value

    return d


def get_rows_by_indices(self, index1, index2) -> dict:
    row1 = self.get_row_by_index(index1)
    row2 = self.get_row_by_index(index2)

    assert row1['cell_id'] == row2['cell_id'], (row1['cell_id'], row2['cell_id'])

    return {
        **{f'{Chain.get_chain(row1["chain"]).name.lower()}_{key}': val for key, val in row1.items()},
        **{f'{Chain.get_chain(row2["chain"]).name.lower()}_{key}': val for key, val in row2.items()},
    }


def get_single_row_value(self, attr_name: str):
    if hasattr(self, attr_name) and getattr(self, attr_name) is not None:
        val = getattr(self, attr_name)
        if isinstance(val, EncodedArray):
            return val.to_string()
        elif isinstance(val, np.ndarray):
            return val.item()
    else:
        return None


def make_element_dataset_objects(bnp_object, class_name) -> list:
    cls = ReflectionHandler.get_class_by_name(class_name, 'data_model')
    if class_name == 'ReceptorSequence':
        objects = [cls.create_from_record(**bnp_object.get_row_by_index(i)) for i in range(len(bnp_object))]
    else:
        objects = [cls.create_from_record(**bnp_object.get_rows_by_indices(i, i + 1)) for i in
                   range(0, len(bnp_object), 2)]
    return objects


def make_buffer_type_from_dataset_file(dataset_file: Path):
    if dataset_file is not None and dataset_file.exists() and dataset_file.is_file():
        with dataset_file.open('r') as file:
            metadata = yaml.safe_load(file)

        type_dict = {key: SequenceSet.STR_TO_TYPE[val] for key, val in metadata["type_dict"].items()}
        dataclass = make_dynamic_seq_set_dataclass(type_dict)
        return bnp.io.delimited_buffers.get_bufferclass_for_datatype(dataclass, delimiter='\t', has_header=True)
    else:
        raise RuntimeError(f"Dataset file {dataset_file} doesn't exist, cannot load the dataset.")


def merge_dataclass_objects(objects: list):
    field_names = sorted(
        list(set(chain.from_iterable([field.name for field in get_fields(obj)] for obj in objects))))

    for obj in objects:
        assert all(hasattr(obj, field) for field in field_names), (obj, field_names)

    cls = type(objects[0])
    return cls(
        **{field_name: list(chain.from_iterable([getattr(obj, field_name) for obj in objects])) for field_name in
           field_names})



import copy
from pathlib import Path

import yaml

from immuneML.app.ImmuneMLApp import ImmuneMLApp
from immuneML.dsl.definition_parsers.ReportParser import ReportParser
from immuneML.dsl.symbol_table.SymbolTable import SymbolTable
from immuneML.dsl.symbol_table.SymbolType import SymbolType
from immuneML.presentation.html.MultiDatasetBenchmarkHTMLBuilder import MultiDatasetBenchmarkHTMLBuilder
from immuneML.util.Logger import print_log
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class MultiDatasetBenchmarkTool:
    """
    MultiDatasetBenchmarkTool trains the models using nested cross-validation (CV) to determine optimal model on multiple datasets. Internally, it uses
    TrainMLModel instruction for each of the listed datasets and performs nested CV on each, accumulates the results of these runs and then
    generates reports on the cumulative results.

    YAML specification:

    .. highlight:: yaml
    .. code-block:: yaml

        definitions: # everything under definitions can be defined in a standard way
            datasets:
                d1: ...
                d2: ...
                d3: ...
            ml_methods:
                ml1: ...
                ml2: ...
                ml3: ...
            encodings:
                enc1: ...
                enc2: ...
            reports:
                r1: ...
                r2: ...
        instructions: # there can be only one instruction
            benchmark_instruction:
                type: TrainMLModel
                benchmark_reports: [r1, r2] # list of reports that will be executed on the results for all datasets
                datasets: [d1, d2, d3] # the same optimization will be performed separately for each dataset
                settings: # a list of combinations of preprocessing, encoding and ml_method to optimize over
                - encoding: enc1 # mandatory field
                  ml_method: ml1 # mandatory field
                - encoding: enc2
                  ml_method: ml2
                - encoding: enc2
                  ml_method: ml3
                assessment: # outer loop of nested CV
                    split_strategy: random # perform Monte Carlo CV (randomly split the data into train and test)
                    split_count: 1 # how many train/test datasets to generate
                    training_percentage: 0.7 # what percentage of the original data should be used for the training set
                selection: # inner loop of nested CV
                    split_strategy: k_fold # perform k-fold CV
                    split_count: 5 # how many fold to create: here these two parameters mean: do 5-fold CV
                labels: # list of labels to optimize the classifier for, as given in the metadata for the dataset
                    - celiac
                strategy: GridSearch # how to choose the combinations which to test from settings (GridSearch means test all)
                metrics: # list of metrics to compute for all settings, but these do not influence the choice of optimal model
                    - accuracy
                    - auc
                reports: # reports to execute on the dataset (before CV, splitting, encoding etc.)
                    - rep1
                number_of_processes: 4 # number of parallel processes to create (could speed up the computation)
                optimization_metric: balanced_accuracy # the metric to use for choosing the optimal model and during training

    """

    def __init__(self, specification_path: Path, result_path: Path, **kwargs):
        self.specification_path = specification_path
        self.result_path = result_path
        self.reports = None

    def run(self):
        print_log("Starting MultiDatasetBenchmarkTool...")
        PathBuilder.build(self.result_path)
        specs = self._split_specs_file()
        self._extract_reports()
        instruction_states = {}
        for index, specs_name in enumerate(specs.keys()):
            print_log(f"Running nested cross-validation on dataset {specs_name} ({index+1}/{len(list(specs.keys()))})..")
            app = ImmuneMLApp(specification_path=specs[specs_name], result_path=self.result_path / specs_name)
            instruction_states[specs_name] = app.run()[0]
            print_log(f"Finished nested cross-validation on dataset {specs_name} ({index+1}/{len(list(specs.keys()))})..")

        print_log("Running reports on the results of nested cross-validation on all datasets...")
        report_results = self._run_reports(instruction_states)
        print_log("Finished reports, now generating HTML output...")
        MultiDatasetBenchmarkHTMLBuilder.build(report_results, self.result_path,
                                               {specs_name: self.result_path / specs_name for specs_name in specs.keys()})
        print_log("MultiDatasetBenchmarkTool finished.")

    def _extract_reports(self):
        with self.specification_path.open("r") as file:
            workflow_specification = yaml.safe_load(file)

        report_keys = list(workflow_specification['instructions'].values())[0]['benchmark_reports']

        ParameterValidator.assert_all_in_valid_list(report_keys, list(workflow_specification['definitions']['reports'].keys()),
                                                    MultiDatasetBenchmarkTool.__name__, "benchmark_reports")

        reports = {key: value for key, value in workflow_specification['definitions']['reports'].items() if key in report_keys}
        symbol_table, _ = ReportParser.parse(reports, SymbolTable())
        self.reports = [entry.item for entry in symbol_table.get_by_type(SymbolType.REPORT)]

    def _split_specs_file(self) -> dict:
        with self.specification_path.open("r") as file:
            workflow_specification = yaml.safe_load(file)

        self._check_specs(workflow_specification)

        specs_files = {}

        instruction_name = list(workflow_specification['instructions'].keys())[0]
        instruction = workflow_specification['instructions'][instruction_name]

        for dataset_name in instruction['datasets']:
            new_specs = copy.deepcopy(workflow_specification)
            new_specs['definitions']['datasets'] = {dataset_name: new_specs['definitions']['datasets'][dataset_name]}
            del new_specs['instructions'][instruction_name]['datasets']
            del new_specs['instructions'][instruction_name]['benchmark_reports']
            new_specs['instructions'][instruction_name]['dataset'] = dataset_name
            new_specs_file = self.result_path / f"specs_{dataset_name}.yaml"
            with new_specs_file.open('w') as file:
                yaml.dump(new_specs, file)
            specs_files[dataset_name] = new_specs_file

        return specs_files

    def _check_specs(self, workflow_specification):
        location = 'MultiDatasetBenchmarkTool'
        ParameterValidator.assert_keys(workflow_specification.keys(), ['definitions', 'instructions', 'output'], location, 'YAML specification')

        self._check_dataset_specs(workflow_specification, location)
        self._check_instruction_specs(workflow_specification, location)

    def _check_dataset_specs(self, workflow_specification, location):
        ParameterValidator.assert_type_and_value(workflow_specification['definitions'], dict, location, 'definitions')
        ParameterValidator.assert_keys_present(workflow_specification['definitions'].keys(), ['datasets'], location, 'definitions')
        ParameterValidator.assert_type_and_value(workflow_specification['definitions']['datasets'], dict, location, 'datasets')

        dataset_names = list(workflow_specification['definitions']['datasets'].keys())

        assert len(dataset_names) > 1, \
            f"MultiDatasetBenchmarkTool: there is only one dataset specified ({dataset_names[0]}), while this tool operates on multiple datasets. " \
            f"If only one dataset is needed, consider using the training instruction directly."

    def _check_instruction_specs(self, workflow_specification, location):
        ParameterValidator.assert_type_and_value(workflow_specification['instructions'], dict, location, 'instructions')

        instruction_names = list(workflow_specification['instructions'].keys())
        assert len(instruction_names) == 1, f"MultiDatasetBenchmarkTool: there can be only one instruction specified for this tool. " \
                                            f"Currently the following instructions are specified: {instruction_names}."

        ParameterValidator.assert_keys_present(workflow_specification['instructions'][instruction_names[0]].keys(), ['type', 'datasets'], location,
                                               instruction_names[0])

        instruction_type = workflow_specification['instructions'][instruction_names[0]]['type']
        assert instruction_type == 'TrainMLModel', \
            f"MultiDatasetBenchmarkTool: this tool works only with instruction of type 'TrainMLModel', got {instruction_type} instead."

        datasets_in_instruction = workflow_specification['instructions'][instruction_names[0]]['datasets']
        assert len(datasets_in_instruction) > 1, \
            f'{location}: this tool takes a multiple dataset names as input, but only {len(datasets_in_instruction)} were provided: ' \
            f'{datasets_in_instruction}.'

    def _run_reports(self, instruction_states: dict):
        report_results = {}
        for index, report in enumerate(self.reports):
            print_log(f"Running report {report.name} ({index+1}/{len(self.reports)})...")
            report.instruction_states = list(instruction_states.values())
            report.result_path = PathBuilder.build(self.result_path / 'benchmarking_reports/')
            report_result = report.generate_report()
            report_results[report.name] = report_result

        return report_results



import logging
from abc import ABC

import pandas as pd

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.data_model.dataset.RepertoireDataset import RepertoireDataset
from immuneML.preprocessing.Preprocessor import Preprocessor
from immuneML.util.PathBuilder import PathBuilder


class Filter(Preprocessor, ABC):

    def _build_new_metadata(self, dataset: RepertoireDataset, indices_to_keep: list):
        if dataset.metadata_file:
            df = pd.read_csv(dataset.metadata_file).iloc[indices_to_keep, :]
            df.reset_index(drop=True, inplace=True)

            PathBuilder.build(self.result_path)
            path = self.result_path / f"{dataset.metadata_file.stem}_filtered.csv"
            df.to_csv(path, index=False)
        else:
            path = None
        return path

    def _remove_empty_repertoires(self, repertoires: list):
        filtered_repertoires = []
        removed_repertoire_info = []
        for repertoire in repertoires:
            if len(repertoire.sequences) > 0:
                filtered_repertoires.append(repertoire)
            else:
                removed_repertoire_info.append({"id": repertoire.identifier,
                                                'subject_id': repertoire.metadata['subject_id']
                                                if repertoire.metadata is not None
                                                   and 'subject_id' in repertoire.metadata else ''})

        logging.info(f"Removed {len(removed_repertoire_info)} repertoires:\n{removed_repertoire_info}")
        return filtered_repertoires

    def check_dataset_not_empty(self, processed_dataset: Dataset, location="Filter"):
        assert processed_dataset.get_example_count() > 0, f"{location}: {type(processed_dataset).__name__} ended up empty after filtering. " \
                                                          f"Please adjust filter settings."




import copy
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Dict, Tuple

import pandas as pd
import sklearn

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.ml_metrics.ClusteringMetric import is_internal, is_external
from immuneML.reports.Report import Report
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.clustering_reports.ClusteringReport import ClusteringReport
from immuneML.reports.encoding_reports.EncodingReport import EncodingReport
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.Instruction import Instruction
from immuneML.workflows.instructions.clustering.clustering_run_model import ClusteringSetting, ClusteringItem
from immuneML.workflows.steps.DataEncoder import DataEncoder
from immuneML.workflows.steps.DataEncoderParams import DataEncoderParams


@dataclass
class ClusteringState:
    name: str
    dataset: Dataset
    metrics: List[str]
    clustering_settings: List[ClusteringSetting]
    clustering_items: List[ClusteringItem] = field(default_factory=list)
    result_path: Path = None
    label_config: LabelConfiguration = None
    predictions_path: Path = None
    cl_item_report_results: Dict[str, Dict[str, List[ReportResult]]] = None
    clustering_report_results: List[ReportResult] = field(default_factory=list)


class ClusteringInstruction(Instruction):
    """
    Clustering instruction fits clustering methods to the provided encoded dataset and compares the combinations of
    clustering method with its hyperparameters, and encodings across a pre-defined set of metrics. Finally, it
    provides options to include a set of reports to visualize the results.

    .. note::

        This is an experimental feature in version 3.0.0a1.

    Specification arguments:

    - dataset (str): name of the dataset to be clustered

    - metrics (list): a list of metrics to use for comparison of clustering algorithms and encodings (it can include
      metrics for either internal evaluation if no labels are provided or metrics for external evaluation so that the
      clusters can be compared against a list of predefined labels)

    - labels (list): an optional list of labels to use for external evaluation of clustering

    - clustering_settings (list): a list of combinations of encoding, optional dimensionality reduction algorithm, and
      the clustering algorithm that will be evaluated

    - reports (list): a list of reports to be run on the clustering results or the encoded data

    - number_of_processes (int): how many processes to use for parallelization

    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_clustering_instruction:
            type: Clustering
            dataset: d1
            metrics: [adjusted_rand_score, adjusted_mutual_info_score]
            labels: [epitope, v_call]
            clustering_settings:
                - encoding: e1
                  dim_reduction: pca
                  method: k_means1
                - encoding: e2
                  method: dbscan
            reports: [rep1, rep2]

    """

    def __init__(self, dataset: Dataset, metrics: List[str], clustering_settings: List[ClusteringSetting],
                 name: str, label_config: LabelConfiguration = None, reports: List[Report] = None,
                 number_of_processes: int = None):
        self.state = ClusteringState(name, dataset, metrics, clustering_settings, label_config=label_config)
        self.reports = reports
        self.number_of_processes = number_of_processes

    def run(self, result_path: Path):
        self._setup_paths(result_path)
        self._init_report_result_structure()

        predictions_df = self.state.dataset.get_metadata(self.state.label_config.get_labels_by_name(), return_df=True)
        predictions_df['example_id'] = self.state.dataset.get_example_ids()

        for cl_setting in self.state.clustering_settings:
            cl_item, predictions_df = self._run_clustering_setting(cl_setting, predictions_df)
            self.state.clustering_items.append(cl_item)

        predictions_df.to_csv(self.state.predictions_path, index=False)
        self._run_clustering_reports()

        return self.state

    def _setup_paths(self, result_path: Path):
        self.state.result_path = PathBuilder.build(result_path)
        self.state.predictions_path = self.state.result_path / 'predictions.csv'

    def _run_clustering_setting(self, cl_setting: ClusteringSetting, predictions_df: pd.DataFrame) \
            -> Tuple[ClusteringItem, pd.DataFrame]:

        cl_setting_path = PathBuilder.build(self.state.result_path / cl_setting.get_key())

        dataset = DataEncoder.run(DataEncoderParams(dataset=self.state.dataset, encoder=cl_setting.encoder,
                                                    encoder_params=EncoderParams(model=cl_setting.encoder_params,
                                                                                 result_path=cl_setting_path,
                                                                                 pool_size=self.number_of_processes,
                                                                                 label_config=self.state.label_config,
                                                                                 learn_model=True,
                                                                                 encode_labels=False)))
        if cl_setting.dim_reduction_method is not None:
            dataset.encoded_data.dimensionality_reduced_data = cl_setting.dim_reduction_method.fit_transform(dataset)

        cl_setting.clustering_method.fit(dataset)
        predictions = cl_setting.clustering_method.predict(dataset)
        predictions_df[f'predictions_{cl_setting.get_key()}'] = predictions

        features = dataset.encoded_data.examples if cl_setting.dim_reduction_method is None \
            else dataset.encoded_data.dimensionality_reduced_data
        ext_performances, int_performances = self._evaluate_clustering(predictions_df, cl_setting.get_key(), features)

        cl_item = ClusteringItem(cl_setting=cl_setting, dataset=dataset, predictions=predictions,
                                 performance=None)

        # performances.to_csv(str(cl_setting_path / 'performances.csv'), index=False)
        self._run_reports(cl_item)

        cl_item.dataset.encoded_data = None

        return cl_item, predictions_df

    def _evaluate_clustering(self, predictions: pd.DataFrame, cl_setting_name: str, features) \
            -> Tuple[pd.DataFrame, pd.DataFrame]:
        if self.state.label_config is not None and self.state.label_config.get_label_count() > 0:

            external_performances = {label: {} for label in self.state.label_config.get_labels_by_name()}
            internal_performances = {}

            for metric in [m for m in self.state.metrics if is_internal(m)]:
                metric_fn = getattr(sklearn.metrics, metric)
                internal_performances[metric] = metric_fn(features, predictions[f'predictions_{cl_setting_name}'].values)

            for metric in [m for m in self.state.metrics if is_external(m)]:
                metric_fn = getattr(sklearn.metrics, metric)
                for label in self.state.label_config.get_labels_by_name():
                    external_performances[label][metric] = metric_fn(predictions[label].values,
                                                                     predictions[
                                                                         f'predictions_{cl_setting_name}'].values)

            return (pd.DataFrame(external_performances).reset_index().rename(columns={'index': 'metric'}),
                    pd.DataFrame(internal_performances))

    def _run_clustering_reports(self):
        report_path = PathBuilder.build(self.state.result_path / f'reports/')
        for report in self.reports:
            if isinstance(report, ClusteringReport):
                tmp_report = copy.deepcopy(report)
                tmp_report.result_path = report_path
                tmp_report.cl_state = self.state
                self.state.clustering_report_results.append(tmp_report.generate_report())

        if len(self.reports) > 0:
            gen_rep_count = len(self.state.clustering_report_results)
            print_log(f"{self.state.name}: generated {gen_rep_count} clustering reports.", True)

    def _run_reports(self, cl_item: ClusteringItem):
        report_path = PathBuilder.build(self.state.result_path / f'reports/{cl_item.cl_setting.get_key()}')
        for report in self.reports:
            if isinstance(report, EncodingReport):
                tmp_report = copy.deepcopy(report)
                tmp_report.result_path = report_path
                tmp_report.dataset = cl_item.dataset
                rep_result = tmp_report.generate_report()
                self.state.cl_item_report_results[cl_item.cl_setting.get_key()]['encoding'].append(rep_result)

        if len(self.reports) > 0:
            gen_rep_count = sum(len(reports) for rep_type, reports in
                                self.state.cl_item_report_results[cl_item.cl_setting.get_key()].items())
            print_log(f"{self.state.name}: generated {gen_rep_count} reports for setting "
                      f"{cl_item.cl_setting.get_key()}.", True)

    def _init_report_result_structure(self):
        self.state.cl_item_report_results = {
            cl_setting.get_key(): {'encoding': []} for cl_setting in self.state.clustering_settings
        }
